from pydtnn.activations.activation import Activation
from pydtnn.tracers.events import PYDTNN_MDL_EVENT, PYDTNN_MDL_EVENTS, PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, \
    PYDTNN_EVENT_FINISHED, PYDTNN_MDL_EVENT_enum, PYDTNN_OPS_EVENT_enum

try:
    from pydtnn.libs.libmpi import MPI
except Exception:
    pass
import numpy as np


class ActivationCPU(Activation[np.ndarray]):
    """
    Extends an Activation class with the attributes and methods required by CPU Activations.

    The next methods are copied from LayerCPU:
      * reduce_weights_async()
      * wait_allreduce_async()
      * reduce_weights_sync()
    """

    def initialize(self, prev_shape, x: np.ndarray | None = None):
        super().initialize(prev_shape, x)

    def reduce_weights_async(self, gradient: bool = True) -> None:
        if not self.model.comm:
            return
        self.reqs_allred = {}

        for w_, dw_ in self.grad_vars.items():
            dw_ = dw_ if gradient else w_
            dw: np.ndarray = getattr(self, dw_)
            np.multiply(dw, self.model.rank_weight, out=dw,
                        dtype=self.model.dtype)
            if self.model.crypt:
                dw = self.model.crypt.encrypt(dw)
            if self.model.use_mpi_buffers:
                req = self.model.comm.Iallreduce(MPI.IN_PLACE, dw, op=MPI.SUM)
            else:
                req = self.model.comm.iallreduce(dw, op=MPI.SUM)
            self.reqs_allred[dw_] = req

    def wait_allreduce_async(self, gradient=True) -> None:
        if not self.model.comm or self.model.enable_nccl:
            return
        for w_, dw_ in self.grad_vars.items():
            dw_ = dw_ if gradient else w_
            res = self.reqs_allred[dw_].wait()
            if res is None:
                dw = getattr(self, dw_)
            else:
                dw = res
            if self.model.crypt:
                dw = self.model.crypt.decrypt(dw)
            setattr(self, dw_, dw)

    def reduce_weights_sync(self, gradient=True) -> None:
        if not self.model.comm:
            return
        for w_, dw_ in self.grad_vars.items():
            dw_ = dw_ if gradient else w_
            self.model.tracer.emit_nevent([PYDTNN_MDL_EVENT, PYDTNN_OPS_EVENT],
                                          [self.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.ALLREDUCE_DW,
                                           self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.OPS_ALLREDUCE_DW])
            dw: np.ndarray = getattr(self, dw_)
            np.multiply(dw, self.model.rank_weight, out=dw,
                        dtype=self.model.dtype)
            if self.model.crypt:
                dw = self.model.crypt.encrypt(dw)  #type: ignore
            if self.model.use_mpi_buffers:
                self.model.comm.Allreduce(MPI.IN_PLACE, dw, op=MPI.SUM)
            else:
                dw = self.model.comm.allreduce(dw, op=MPI.SUM)
            if self.model.crypt:
                dw = self.model.crypt.decrypt(dw)  #type: ignore
            setattr(self, dw_, dw)
            self.model.tracer.emit_nevent([PYDTNN_MDL_EVENT, PYDTNN_OPS_EVENT], [PYDTNN_EVENT_FINISHED, PYDTNN_EVENT_FINISHED])



import time
import itertools

import numpy as np

from pydtnn.tests.conv_2d_conv_gemm import D, Conv2DConvGemmTestCase
from pydtnn.utils import random
from pydtnn.libs.libmpi import MPI


class Conv2DConvGemmSlowTestCase(Conv2DConvGemmTestCase):
    """
    Tests that Conv2D with conv_gemm leads to the same results than Conv2d with mm and i2c.T (exhaustive version)
    """
    # NOTE: Delete parent test to prevent re-export and re-testing
    global Conv2DConvGemmTestCase
    del Conv2DConvGemmTestCase

    random.seed(0)  # type: ignore
    dtype = np.float32

    R = list(itertools.product(
        range(1, 4),   # kn
        range(1, 4),   # b
        range(8, 11),  # c
        range(8, 11),  # h
        range(8, 11),  # w
        range(2, 12),  # kh
        range(2, 12),  # kw
        range(0, 4),   # vp
        range(0, 4),   # hp
        range(1, 4),   # vs
        range(1, 4)    # hs
    ))

    X = random.random((
        4,   # b
        11,  # c
        11,  # h
        11   # w
    )).astype(dtype, order="C")

    W = random.random((
        4,   # kn
        11,  # c
        11,  # kh
        11,  # kw
    )).astype(dtype, order="C")

    def test_forward_backward_multiple_params(self):
        """Tests that different input matrices, paddings and strides, lead to the same solution"""
        start = time.time()

        comm = MPI.COMM_WORLD
        batch = self.R[comm.rank::comm.size]

        for i, params in enumerate(batch):
            self._test_forward_backward_multiple_params(*params)

            if comm.rank:
                continue

            i *= comm.size
            perc = (i + 1) / len(self.R)
            elapsed = time.time() - start
            remain = (1 - perc) * (elapsed / perc)
            m, s = divmod(remain, 60)
            h, m = divmod(m, 60)
            d, h = divmod(h, 24)
            print(f"{perc:.5%} (eta: {d:2.0f}d {h:2.0f}h {m:2.0f}m {s:2.0f}s)", end="\r")

    def _test_forward_backward_multiple_params(self, kn: int, b: int, c: int, h: int, w: int, kh: int, kw: int, vpadding: int, hpadding: int, vstride: int, hstride: int):
        d = D(kn=kn, b=b, c=c, h=h, w=w, kh=kh, kw=kw, vpadding=vpadding, hpadding=hpadding, vstride=vstride, hstride=hstride)

        if d.kh >= d.h + 1 or d.kw >= d.w + 1:
            return

        if d.b != 1 or d.c != 1:
            x = self.X[:d.b, :d.c, :d.h, :d.w].copy(order="C")
        else:
            x = np.concatenate([
                np.arange(b := (i + 1) * 100, b + d.w, dtype=np.float32)
                for i in range(d.h)
            ]).reshape((d.b, d.c, d.h, d.w))

        weights = self.W[:d.kn, :d.c, :d.kh, :d.kw].copy(order="C")

        self._test_forward_backward(d, x, weights)



from pydtnn.backends.cpu.layers.abstract.conv_2d_standard import Conv2DStandardCPU
from pydtnn.backends.cpu.utils.im2col_nchw_cython import im2col_nchw_cython
from pydtnn.backends.cpu.utils.im2row_nhwc_cython import im2row_nhwc_cython
from pydtnn.libs.libconvwinograd import ConvWinograd
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum

import numpy as np

from pydtnn.utils.tensor import TensorFormat


class Conv2DWinogradCPU(Conv2DStandardCPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # convWinograd related attributes (will be initialized in initialize())
        self.cw: ConvWinograd = None  # type: ignore

    def initialize(self, prev_shape, x: np.ndarray | None = None):
        super().initialize(prev_shape, x)
        # ConvWinograd parameters
        self.cw = ConvWinograd(self.kh, self.kw, self.vstride, self.hstride,
                                self.vdilation, self.hdilation,
                                dtype=self.model.dtype, tensor_format=self.model.tensor_format,
                                debug=self.debug, parent_layer=self)
        
        match self.model.tensor_format:
            case TensorFormat.NCHW:
                self.forward = self._forward_cw_nchw
                self.backward = self._backward_cw_nchw
            case TensorFormat.NHWC:
                self.forward = self._forward_cw_nhwc
                self.backward = self._backward_cw_nhwc
            case _:
                raise NotImplementedError(f"\"{self.model.tensor_format}\" format not implemented.")
        # ---

    def _forward_cw_nhwc(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses the convWinograd library"""

        self.cw_x = x

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONVWINOGRAD)
        y: np.ndarray = self.cw.conv_winograd_nhwc(self.weights, x, biases=self.biases,
                                                vpadding=self.vpadding, hpadding=self.hpadding,
                                                vstride=self.vstride, hstride=self.hstride,
                                                vdilation=self.vdilation, hdilation=self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return y

    def _forward_cw_nchw(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses the convWinograd library"""

        self.cw_x = x

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONVWINOGRAD)
        y: np.ndarray = self.cw.conv_winograd_nchw(self.weights, x, biases=self.biases,
                                                vpadding=self.vpadding, hpadding=self.hpadding,
                                                vstride=self.vstride, hstride=self.hstride,
                                                vdilation=self.vdilation, hdilation=self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return y

    def _backward_cw_nhwc(self, dy: np.ndarray) -> np.ndarray:
        """Version of the backward function that uses the convWinograd library"""

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_IM2COL)

        self.x_rows = np.zeros(((dy.shape[0] * self.ho * self.wo), (self.ci * self.kh * self.kw)), dtype=self.model.dtype)
        im2row_nhwc_cython(self.cw_x, self.x_rows,
                           self.kh, self.kw, self.ho, self.wo,
                           self.vpadding, self.hpadding,
                           self.vstride, self.hstride, self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return self._backward_i2c_nhwc(dy)

    def _backward_cw_nchw(self, dy: np.ndarray) -> np.ndarray:
        """Version of the backward function that uses the convWinograd library"""
        n, c, _, _ = dy.shape
        self.x_cols = np.zeros((c * self.kh * self.kw, n * self.ho * self.wo))
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_IM2COL)
        im2col_nchw_cython(self.cw_x,
                           self.x_cols,
                           self.kh, self.kw, 
                           self.ho, self.wo,
                           self.vpadding, self.hpadding,
                           self.vstride, self.hstride, 
                           self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return self._backward_i2c_nchw(dy)



import numpy as np
from pydtnn.backends.cpu.utils.sigmoid_cython import sigmoid_bwd_cython, sigmoid_fwd_cython

from pydtnn.activations.sigmoid import Sigmoid
from pydtnn.backends.cpu.activations.activation import ActivationCPU


class SigmoidCPU(Sigmoid[np.ndarray], ActivationCPU):

    def initialize(self, prev_shape, x=None):
        super().initialize(prev_shape, x)
        self.y: np.ndarray = None  # type: ignore (the value will be set in forward)

        # NOTE: These attributes only store data, their values before the operation doesn't matter; they're initalized due avoid warnings in "LayerAndActivationBase.export".
        self._y = np.zeros(shape=(self.model.batch_size, *prev_shape), dtype=self.model.dtype, order="C")
        self.dx = np.zeros(shape=(self.model.batch_size, *prev_shape), dtype=self.model.dtype, order="C")

    def forward(self, x: np.ndarray) -> np.ndarray:

        self.y = self._y[:x.shape[0], :]
        sigmoid_fwd_cython(x.reshape(-1, copy=False), self.y.reshape(-1, copy=False))
        return self.y

    def backward(self, dy: np.ndarray) -> np.ndarray:

        dx = self.dx[:dy.shape[0], :]
        sigmoid_bwd_cython(dy.reshape(-1, copy=False, order="C"),
                           self.y.reshape(-1, copy=False, order="C"),
                           dx.reshape(-1, copy=False, order="C"))

        return dx



import numpy as np

from pydtnn.activations.softmax import Softmax
from pydtnn.backends.cpu.activations.activation import ActivationCPU
from pydtnn.utils.constants import ArrayShape

class SoftmaxCPU(Softmax[np.ndarray], ActivationCPU):
    def initialize(self, prev_shape, x=None):
        super().initialize(prev_shape, x)
        self.y: np.ndarray

        shape_intermediate_ops = list(self.shape)
        shape_intermediate_ops[self.axis_dim - 1] = 1

        # NOTE: These attributes only store data, their value before the operation doesn't matter; they're initalized due avoid warnings in "LayerAndActivationBase.export".
        self._y = np.zeros(shape=(self.model.batch_size, *self.shape),
                           dtype=self.model.dtype, order="C")
        self.mul_dy = np.zeros(shape=(self.model.batch_size, *self.shape),
                               dtype=self.model.dtype, order="C")
        self.max_x = np.zeros(shape=(self.model.batch_size, *shape_intermediate_ops),
                              dtype=self.model.dtype, order="C")
        self.sum_y = np.zeros(shape=(self.model.batch_size, *shape_intermediate_ops),
                              dtype=self.model.dtype, order="C")
        self.sum_dy = np.zeros(shape=(self.model.batch_size, *shape_intermediate_ops),
                               dtype=self.model.dtype, order="C")

    def forward(self, x: np.ndarray) -> np.ndarray:
        # self.y = np.exp(x - np.max(x, axis=1, keepdims=True))
        # self.y /= np.sum(self.y, axis=1, keepdims=True)
        # return self.y
        self.y = self._y[:x.shape[0], :]
        max_x = self.max_x[:x.shape[0], :]
        sum_y = self.sum_y[:x.shape[0], :]

        np.max(x, axis=self.axis_dim, keepdims=True, out=max_x)
        np.subtract(x, max_x, out=x,
                    dtype=self.model.dtype)
        np.exp(x, out=self.y,
               dtype=self.model.dtype)
        np.sum(self.y, axis=self.axis_dim, keepdims=True, out=sum_y)
        np.divide(self.y, sum_y, out=self.y,
                  dtype=self.model.dtype, order="C")
        return self.y

    def backward(self, dy: np.ndarray) -> np.ndarray:
        # return self.y * (dy - (dy * self.y).sum(axis=1, keepdims=True))
        sum_dy = self.sum_dy[:dy.shape[0], :]
        mul_dy = self.mul_dy[:dy.shape[0], :]

        np.multiply(dy, self.y, out=mul_dy,
                    dtype=self.model.dtype)
        mul_dy.sum(axis=self.axis_dim, keepdims=True, out=sum_dy)
        np.subtract(dy, sum_dy, out=dy,
                    dtype=self.model.dtype)
        np.multiply(self.y, dy, out=dy,
                    dtype=self.model.dtype, order="C")

        return dy



from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.model import Model
from pydtnn.backends.cpu.layers.conv_2d import Conv2DCPU
from pydtnn.tests.abstract.common import D
from pydtnn.tests.abstract.common import Params
from pydtnn.tests.abstract.conv_2d_common import Conv2DCommonTestCase
from pydtnn.utils.tensor import TensorFormat
from pydtnn.utils.initializers import glorot_uniform, zeros


class Conv2DGroupTestCase(Conv2DCommonTestCase):
    """
    Tests that Conv2D with Depth+Pair leads to the same results than Conv2D Standard
    """
    # NOTE: Delete parent test to prevent re-export and re-testing
    global Conv2DCommonTestCase
    del Conv2DCommonTestCase

    @staticmethod
    def _get_layers(d: D, deconv=False, trans=False) -> tuple[Conv2DCPU, Conv2DCPU]:
        params = Params()
        params.tensor_format = TensorFormat.NHWC.upper()
        params.batch_size = d.b
        model = Model(**vars(params))
        model.mode = Model.Mode.TRAIN

        conv2d_depth = Conv2D(nfilters=d.kn, filter_shape=(d.kh, d.kw),
                              grouping=Conv2D.Grouping.DEPTHWISE,
                              padding=(d.vpadding, d.hpadding),
                              stride=(d.vstride, d.hstride),
                              dilation=(d.vdilation, d.hdilation),
                              use_bias=True, weights_initializer=glorot_uniform, biases_initializer=zeros)
        conv2d_pair = Conv2D(nfilters=d.kn, filter_shape=(d.kh, d.kw),
                             grouping=Conv2D.Grouping.POINTWISE,
                             padding=(d.vpadding, d.hpadding),
                             stride=(d.vstride, d.hstride),
                             dilation=(d.vdilation, d.hdilation),
                             use_bias=True, weights_initializer=glorot_uniform, biases_initializer=zeros)
        chain = ConcatenationBlock([
            conv2d_depth,
            conv2d_pair
        ])
        chain.init_backend_from_model(model)
        chain.initialize(prev_shape=(d.c, d.h, d.w), x=None)

        conv2d = Conv2D(nfilters=d.kn, filter_shape=(d.kh, d.kw),
                        grouping=Conv2D.Grouping.STANDARD,
                        padding=(d.vpadding, d.hpadding),
                        stride=(d.vstride, d.hstride),
                        dilation=(d.vdilation, d.hdilation),
                        use_bias=True, weights_initializer=glorot_uniform, biases_initializer=zeros)
        conv2d.init_backend_from_model(model)
        conv2d.initialize(prev_shape=(d.c, d.h, d.w), x=None)

        # Set the same initial weights and biases to both layers
        conv2d_depth.weights = conv2d.weights.copy()
        conv2d_depth.biases = conv2d.biases.copy()
        conv2d_pair.weights = conv2d.weights.copy()
        conv2d_pair.biases = conv2d.biases.copy()

        return conv2d, chain

    @staticmethod
    def _set_state(layer: Conv2D, weights) -> None:
        if isinstance(layer, ConcatenationBlock):
            layer.paths[0][0].weights = weights.copy()
            layer.paths[0][1].weights = weights.copy()
        else:
            layer.weights = weights.copy()



import ctypes
import os
from importlib import import_module

from pydtnn.tracers.tracer import Tracer


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pydtnn.model import Model


class ExtraeTracer(Tracer):
    """
    ExtraTracer
    """

    def __init__(self, tracing: bool):
        super().__init__(tracing)
        self.pyextrae = None  # Declared here, will be initialized on enable_tracing()

    def enable_tracing(self):
        super().enable_tracing()
        self.pyextrae = import_module('pyextrae.common.extrae')

    def _define_event_types(self, model: "Model"):
        """This method will be called only if tracing is enabled"""
        super()._define_event_types(model)
        for event_type_value, event_type in self.event_types.items():
            description = event_type.name
            nvalues = len(event_type)
            values = (ctypes.c_ulonglong * nvalues)()
            descriptions = (ctypes.c_char_p * nvalues)()
            for i, description in event_type.items():
                values[i] = i
                descriptions[i] = description
            self.pyextrae.Extrae[os.getpid()].Extrae_define_event_type(
                ctypes.pointer(ctypes.c_uint(event_type_value)),
                ctypes.c_char_p(description.encode('utf-8')),
                ctypes.pointer(ctypes.c_uint(nvalues)),
                ctypes.pointer(values),
                ctypes.pointer(descriptions))

    def _emit_event(self, evt_type: int, val: int, stream=None):
        """This method will be called only if tracing is enabled"""
        self.pyextrae.eventandcounters(evt_type, val)

    def _emit_nevent(self, evt: int, val: int, stream=None):
        """This method will be called only if tracing is enabled"""
        self.pyextrae.neventandcounters(evt, val)



from collections.abc import Sequence

from pydtnn.activations.relu import Relu
from pydtnn.activations.sigmoid import Sigmoid
from pydtnn.layer_base import LayerBase
from pydtnn.layers.average_pool_2d import AveragePool2D
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layers.max_pool_2d import MaxPool2D
from pydtnn.utils.initializers import he_uniform


def tsunamis_eflows_UMA(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=input_shape))
    _(Conv2D(nfilters=32, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform))
    _(Conv2D(nfilters=32, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform))

    UMA_blocks = [[32, 64, 64],
                  [64, 128, 128],
                  [128, 256, 256]]

    for n3x3, n3x3red, n2x2 in UMA_blocks:
        _(ConcatenationBlock(
            [Conv2D(nfilters=n3x3, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform),
             Conv2D(nfilters=n3x3red, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform),
             MaxPool2D(pool_shape=(3, 3), stride=2, padding=1)
             ],
            [Conv2D(nfilters=n2x2, filter_shape=(3, 3), padding=1, stride=2, weights_initializer=he_uniform)
             ]))

    UMA_dense_blocks = [[128, 256, 512],
                        [128, 256, 512]]

    for n3x3, n3x3red, n3x3fin in UMA_dense_blocks:
        _(ConcatenationBlock(
            [Conv2D(nfilters=n3x3, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform),
             Conv2D(nfilters=n3x3red, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform),
             Conv2D(nfilters=n3x3fin, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform)
             ], []))

    _(AveragePool2D(pool_shape=(10, 10), stride=1))  # Global average pooling 2D
    _(Flatten())

    for fc in range(3):
        _(FC(shape=(500,), activation=Relu))

    _(FC(shape=output_shape, activation=Sigmoid))

    return model



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def relu_cython[T:_npDT](x: _npDT_1Dims[T],
                         max: _npDT_1Dims[T],
                         mask: _np.ndarray[tuple[int], _np.int8]) -> None:
    """
    Args:
        x (npDT_1Dims): 1-dimensional input's array.
        max (npDT_1Dims): 1-dimensional array where the ouput is stored
        mask (np.ndarray[tuple[int], np.int8]): 1-dimensional array where the output's mask is stored.
    Returns:
        Nothing. The output is stored in "max" and "mask".
    """


def capped_relu_cython[T:_npDT](x: _npDT_1Dims[T],
                                max: _npDT_1Dims[T],
                                mask: _np.ndarray[tuple[int], _np.int8],
                                cap: float) -> None:
    """
    ReLU function where the values above "cap"'s value are set as this value.

    Note: if cap is 6, this is a Relu6

    Args:
        x (npDT_1Dims): 1-dimensional input's array.
        max (npDT_1Dims): 1-dimensional array where the ouput is stored
        mask (np.ndarray[tuple[int], np.int8]): 1-dimensional array where the output's mask is stored.
        cap (float): The ReLU's superior limit. Any value in x greater that this parameter will be set to this parameter in the ouput.
    Returns:
        Nothing. The output is stored in "max" and "mask".
    """


def leaky_relu_cython[T:_npDT](x: _npDT_1Dims[T],
                               max: _npDT_1Dims[T],
                               mask: _npDT_1Dims[T],
                               negative_slope: float) -> None:
    """
    Args:
        x (npDT_1Dims): 1-dimensional input's array.
        max (npDT_1Dims): 1-dimensional array where the ouput is stored
        mask (np.ndarray[tuple[int], np.int8]): 1-dimensional array where the output's mask is stored.
        negative_slope (float): The negative value's multiplayer (if is 0, this function acts as a normal ReLU)
    Returns:
        Nothing. The output is stored in "max" and "mask".
    """



from pydtnn.metrics.metric import Metric
from pydtnn.metrics.binary_confusion_matrix import BinaryConfusionMatrix
from pydtnn.utils.constants import Array


class Recall[T: Array](Metric[T]):
    order = BinaryConfusionMatrix.order + 1
    conf_matrix_metric: BinaryConfusionMatrix = None  # type: ignore
    format = "rec: %.4f"

    def initialize(self) -> None:

        for metric in self.model.metrics_funcs:
            if isinstance(metric, BinaryConfusionMatrix):
                self.conf_matrix_metric = metric
                break
        assert self.conf_matrix_metric is not None, f"Recall requires of {BinaryConfusionMatrix.__name__}"
        super().initialize()



# Typing related
from typing import Any
from pydtnn.layer_base import LayerBase
import numpy as np

# Operations/transformations related
import onnx
from pydtnn.model import Model as PyDTNN_Model
import pydtnn.converters.onnx2pydtnn.constants as cons
from pydtnn.layers.input import Input
from pydtnn.utils.tensor import TensorFormat


# ////////////////////////////////////////////////////
# In order to made some parts of this code, I used other converors' code (specially the "onnx2pytorch" library)
# ////////////////////////////////////////////////////

# Notes:
# All the weights and the data related to the variables is stored in "model_graph.initializer"
#   (Note: "onnx.ModelProto.graph.node" is not a node, it is a list of nodes)
# With _node being an element of onnx.ModelProto.graph.node:
#   _node.input: inputs list. _node.output: outputs list. _node.attribute: list made by all the parameteres and values (they are "AttributeProto")

def extract_shape(data: onnx.ValueInfoProto) -> np.shape:
    # The shape of the inputs/ouputs is more or less a list quite hidden.
    #   NOTE: ONNX allows to have shapes of undefined value, e.g.: (N, 3, 224, 224),
    #       and, if it is not defined, that dimension is stored as 1. It's assumed that it is not relevant data and it will be removed, since "N" are the number of samples.
    #   NOTA:
    #   Parece ser que se guarda como "1" ==> viendo cómo va PyDTNN, lo mejor es que, si detecto que hay una "N" (como la de arriba), eliminarla
    #   ==> Se va a sumir que el primer elemento es el "N" ==> TODO: Hacer esto bien.
    shape = [elem.dim_value for elem in data.type.tensor_type.shape.dim if elem.dim_value != 0]
    del shape[0]
    return tuple(shape)
# --- extract_shape --- #


def get_relevant_data(model_graph: onnx.GraphProto) -> tuple[dict[str, np.shape], dict[str, np.shape], dict[str, np.ndarray]]:

    # onnx.numpy_helper.to_array() is a function that transforms onnx data into a ndarray (numpy's array)

    # Weights dicionary. Key: weight name. Value: the onnx tensor in a numpy format (with -technically- the correct dtype).
    weights_dict = {node.name: onnx.numpy_helper.to_array(node) for node in model_graph.initializer}

    # Inputs dicionary. Key: input name. Value: the shape of the input.
    inputs_dict = {_input.name: extract_shape(_input)
                   for _input in model_graph.input if _input.name not in weights_dict.keys()}

    # Outputs dicionary. Key: output name. Value: the shape of the output.
    outputs_dict = {ouput.name: extract_shape(ouput) for ouput in model_graph.output}

    return (inputs_dict, outputs_dict, weights_dict)
# --- END get_inputs_outputs_and_attributes_names --- #


def extract_attributes(node: onnx.NodeProto) -> dict[str, Any]:

    return {attribute.name: onnx.helper.get_node_attr_value(node, attribute.name)
            for attribute in node.attribute}
# --- END extract_attributes --- #


def get_lists_operations_and_outputs(info: dict[str, Any], operations: dict[str, tuple[LayerBase, list[str]]]) -> tuple[list[list[LayerBase]], list[str]]:

    # NOTE: It is assumed that the model will by a feed-forward network
    dict_branch = {}

    # Making the "path" of layers for every input
    for inpt in info[cons.CONST_INPUTS]:
        dict_branch[inpt] = dict()

        input_search = inpt
        while input_search is not None:
            # operations: {[output_name]: ([operation], [inputs])}
            op, inp = operations[input_search]
            output = input_search

            if inp is None:
                # case: root layer.
                dict_branch[inpt][input_search] = (op, output)
                input_search = None
            else:
                input_search = inp[0]  # The inputs list should have only one input.
                dict_branch[inpt][input_search] = (op, output)

    # Searching the first coincidence

    # Sets are not ordered by insertion ==> keep order with enumerate ==>
    #   ==> braches have different sizes, then the same node may have different order in different branches ==>
    #   ==> that's true from bottom to top, from top to bottom the "intersection layers" (the ones to be searched) should have the same position.
    enumerated_reversed_inputs = enumerate(list(dict_branch[info[cons.CONST_INPUTS][0]].keys())[::-1])

    coincidences = set(enumerated_reversed_inputs)
    for i in range(1, len(info[cons.CONST_INPUTS])):
        coincidences.intersection(set(enumerate(list(dict_branch[info[cons.CONST_INPUTS][i]].keys())[::-1])))

    # "Unenumerating" and sorting the intersection, and getting the first coincidence layer.
    #   ==> NOTE: Due the list was sorting in reverse before, now it is necessary to sort it be reverse again (that's why the "-x[0]").
    coincidence = [elem[1] for elem in sorted(coincidences, key=lambda x: -x[0])][0]

    # Trimming the lists from that element (first coincidence)
    lists_operations = list()
    lists_outputs = list()
    for inpt in info[cons.CONST_INPUTS]:
        _values = list(dict_branch[inpt].values())
        layers = [elem[0] for elem in _values]
        outputs = [elem[1] for elem in _values]

        trimming_index = (layers.index(operations[coincidence][0]))
        lists_operations.append(layers[:trimming_index])  # Remember: list of lists
        lists_outputs.extend(outputs[:trimming_index])  # Remember: list of string

    return (lists_operations, lists_outputs)
# --- END get_lists_operations_and_outputs --- #


def get_actual_inputs(list_inputs: list[str], weights_names: list[str]) -> list[str]:
    # This function' objective is to remove non layer-to-layer onnx inputs (e.g.: the weigth [_weight], the bias [_bias], etc. ).
    #   To do that, only the inputs that end with the accepted ending remains.
    return list(filter(lambda _input: _input not in weights_names, list_inputs))
# --- END get_actual_inputs --- #


def _get_and_put_layer(node: onnx.NodeProto, opset_version: int, operations: dict[str, tuple[LayerBase, list[str]]],
                       weights: dict[str, np.ndarray], output: list[str] | None = None) -> None:

    info = {  # cons.CONST_NODE: node, # Refererence to the model itself (TODO: see if it's necessary. If not ==> delete)
        cons.CONST_OPSET: opset_version,    # Version of the onnx operation
        cons.CONST_INPUTS: get_actual_inputs(list_inputs=node.input, weights_names=list(weights.keys())),   # node's inputs names
        cons.CONST_ALL_INPUTS: node.input,   # ALL node's inputs names (including weights and biases)
        cons.CONST_OUPTUS: node.output if output is None else output,  # node's outputs names or the model's output (TODO: Check if a operation can have multiple outputs)
        cons.CONST_ATTRIBUTES: extract_attributes(node=node),  # dictionary with the node's attributes names and respective values (e.g. the shape of a kernel)
        cons.CONST_WEIGHTS: weights,
        cons.CONST_PREV_LAYERS: operations
    }
    if len(info[cons.CONST_INPUTS]) > 1:
        info[cons.CONST_listS_NODES], operations_to_remove = get_lists_operations_and_outputs(info, operations)
        # Due the way the add/concatention layer works, those must to be removed from the operations (they will be inside the other operation)
        for operation in operations_to_remove:
            del operations[operation]

    operations[info[cons.CONST_OUPTUS][0]] = (cons.switch_onnx_operation_to_pydtnn[node.op_type](info), info[cons.CONST_INPUTS])

    # return Nothing: the output is stored in the dictionary
# --- END _get_and_put_layer --- #


def get_layers(onnx_model: onnx.ModelProto, opset_version: int, inputs: dict[str, np.shape],
               weights: dict[str, np.ndarray], outputs: dict[str, np.shape]) -> list[LayerBase]:

    # TODO: meter otros parámetros que se puedan necesitar
    # operations = list()
    # for node in onnx_model.graph.node:
    #    parameters = [weights[par_name] for par_name in node.input if par_name in weights]
    #    operations.append(SWITCH_ONNX_TO_PYDTNN[node](node, parameters))

    # TODO: implementar las funciones necesarias del "Switch"

    # It is expected to have at least one layer.
    num_operations = len(onnx_model.graph.node)
    assert num_operations > 0

    # operations: {[output_name]: ([operation], [inputs])}
    output_first_layer = get_actual_inputs(list_inputs=onnx_model.graph.node[0].input, weights_names=list(weights.keys()))[0]
    print(f"Input: {inputs[list(inputs.keys())[0]]}")

    # "[None]" indicates that it has no previous layers.
    operations = {output_first_layer: (Input(shape=inputs[list(inputs.keys())[0]]), [None])}

    for i in range(num_operations - 1):
        _get_and_put_layer(node=onnx_model.graph.node[i], opset_version=opset_version, operations=operations, weights=weights)
        print("")  # TODO: Borrar
    _get_and_put_layer(node=onnx_model.graph.node[-1], opset_version=opset_version, operations=operations, weights=weights, output=list(outputs.keys()))

    # The list of layers is returned.
    return list(map(lambda x: x[0], operations.values()))
# --- END get_layers --- #


def load_layers(model: PyDTNN_Model, operations: list[LayerBase]) -> None:

    print(f"Loading layers: {operations}")
    for operation in operations:
        print(f"=> Layer: {operation}")
        print(f"=> Layer shape pre-add: {operation.shape}")
        model.add(operation)
        print(f"=> Layer shape: {operation.shape}")
    print(f"Layers loaded")

    return  # None (No value is returned)
# --- END load_layers --- #


def convert_model(onnx_model: onnx.ModelProto, omm=None, non_blocking_mpi=False, enable_gpu=False, enable_gpudirect=False,
                  enable_nccl=False, dtype=np.float32, tracing=False, tracer_output="", **kwargs) -> PyDTNN_Model:

    if "tensor_format" not in kwargs:
        kwargs["tensor_format"] = TensorFormat.NHWC  # listTensorFormat.NCHW #listTensorFormat.NHWC
    # Output model.
    # NOTE: ¡¡¡¡IMPORTANT!!!!! Be sure that the "parser.model_name" from pydtnn.parser import parser is None!!!!!!!!.
    model = PyDTNN_Model(omm=omm, non_blocking_mpi=non_blocking_mpi, enable_gpu=enable_gpu, enable_gpudirect=enable_gpudirect,
                         enable_nccl=enable_nccl, dtype=dtype, tracing=tracing, tracer_output=tracer_output, **kwargs)

    print("TEST")
    print(model.layers)
    print("TEST")

    # Obtaining the relevant data (inputs, outputs, weights, ...) from the onnx model.
    inputs, outputs, weights = get_relevant_data(onnx_model.graph)
    opset_version = onnx_model.opset_import[0].version

    # Obtaining the operations (layes, activations, etc.).
    operations = get_layers(onnx_model=onnx_model, opset_version=opset_version, inputs=inputs, outputs=outputs, weights=weights)
    # Asigning the operations to the model.
    load_layers(model=model, operations=operations)

    # TODO: Faltaría comprobar el formato de los pesos (y hacer la traducción, si fuese necesario)

    # Loading the weights into the model.
    model.load_store_path(layers=model.layers, d=weights, mode="load")

    return model
# --- END convert_model --- #



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def average_pool_2d_fwd_nchw_cython[T: _npDT](x: _npDT_4Dims[T], y: _npDT_4Dims[T],
                                              kh: int, kw: int, ho: int, wo: int,
                                              vpadding: int, hpadding: int,
                                              vstride: int, hstride: int,
                                              vdilation: int, hdilation: int) -> None:
    """
    Args:
        x (npDT_4Dims): The 4 dimensional input's ndarray.
        y (npDT_4Dims): The 4 dimensional output's ndarray. (the output's data is stored in this parameter).
        kh (int): The kernel's height.
        kw (int): The kernel's width.
        ho (int): The output's height.
        wo (int): The output's width.
        vpadding (int): The vertical padding value.
        hpadding (int): The horizontal padding value.
        vstride (int): The vertical stride value.
        hstride (int): The horizontal stride value.
        vdilation (int): The vertical dilation value.
        hdilation (int): The horizontal dilation value.

    Returns:
        Nothing. Implictily the output is stored in "y".
    """


def average_pool_2d_bwd_nchw_cython[T: _npDT](dy: _npDT_4Dims[T],
                                              dx: _npDT_4Dims[T],
                                              n: int, h: int, w: int, c: int,
                                              kh: int, kw: int, ho: int, wo: int,
                                              vpadding: int, hpadding: int,
                                              vstride: int, hstride: int,
                                              vdilation: int, hdilation: int) -> None:
    """
    Args:
        dy (npDT_4Dims): The 4 dimensional input's ndarray.
        dx (npDT_4Dims): The 4 dimensional output's ndarray. (the output's data will be stored in this parameter). Note: All values in this parameter should be 0.
        n (int): The number of images (usually, the batch size).
        h (int): The images' height.
        w (int): The images' width.
        c (int): The images' number of channel's(e.g.: RGB = 3 channels).
        kh (int): The kernel's height.
        kw (int): The kernel's width.
        ho (int): The output's height.
        wo (int): The output's width.
        vpadding (int): The vertical padding value.
        hpadding (int): The horizontal padding value.
        vstride (int): The vertical stride value.
        hstride (int): The horizontal stride value.
        vdilation (int): The vertical dilation value.
        hdilation (int): The horizontal dilation value.

    Returns:
        Nothing. The output will be stored in "dx".
    """



import numpy as np

from pydtnn.backends.cpu.layers.layer import LayerCPU
from pydtnn.layers.fc import FC
from pydtnn.model import Model
from pydtnn.utils.performance_models import matmul_time
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum


class FCCPU(FC[np.ndarray], LayerCPU):

    biases: np.ndarray

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # The following attributes will be initalized in "initalize"
        self.x: np.ndarray = None  #type: ignore
        self.dw: np.ndarray = None  #type: ignore
        self.db: np.ndarray = None  #type: ignore
    # --

    def initialize(self, prev_shape, x = None):
        super().initialize(prev_shape, x)
        self.weights = self.weights_initializer(self.weights_shape, self.model.dtype)
        # Initialize outputs:
        # NOTE: These attributes only store data, their values before the operation doesn't matter; they're initalized due avoid warnings in "LayerAndActivationBase.export".
        self.db = np.zeros(self.shape, dtype=self.model.dtype, order="C")
        self.dy = np.zeros((self.model.batch_size, *self.shape), dtype=self.model.dtype, order="C")
        self.dw = np.zeros(shape=(*self.prev_shape, *self.shape), dtype=self.model.dtype, order="C")
        self.dx = np.zeros(shape=(self.model.batch_size, *self.prev_shape), dtype=self.model.dtype, order="C")

        if self.use_bias:
            self.biases = self.biases_initializer(self.shape, self.model.dtype)
        self.nparams = self.weights.size + (self.biases.size if self.use_bias else 0)
        # Performance model
        self.fwd_time = \
            matmul_time(m=self.model.batch_size, n=self.weights.shape[1], k=self.weights.shape[0],
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)
        self.bwd_time = \
            matmul_time(m=self.weights.shape[0], n=self.weights.shape[1], k=self.model.batch_size,
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype) + \
            matmul_time(m=self.model.batch_size, n=self.weights.shape[0], k=self.weights.shape[1],
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)  # type: ignore (It works well.)
    # ----

    def forward(self, x: np.ndarray) -> np.ndarray:
        self.x = x
        dy = self.dy[: x.shape[0], :]

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_MATMUL)
        np.matmul(x, self.weights, out=dy, 
                  dtype=self.model.dtype, order="C")

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            np.add(dy, self.biases, out=dy, 
                   dtype=self.model.dtype, order="C")

        return np.asarray(dy, dtype=self.model.dtype, order='C', copy=None)
    # ---

    def backward(self, dy: np.ndarray) -> np.ndarray:

        # self.model.mode = ModelModeEnum.TRAIN is asumed from this point.
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DW_MATMUL)
        # self.dw = np.matmul(self.x.T, dy)
        np.matmul(self.x.T, dy, self.dw, 
                  dtype=self.model.dtype, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            # self.db = np.sum(dy, axis=0)
            np.sum(dy, axis=0, out=self.db)

        dx = self.dx[:self.x.shape[0], :]
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DX_MATMUL)
        # dx = np.matmul(dy, self.weights.T)
        np.matmul(dy, self.weights.T, out=dx, 
                  dtype=self.model.dtype, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(dx, dtype=self.model.dtype, order='C', copy=None)
    # --



import unittest
import warnings

import numpy as np

from pydtnn.layers.addition_block import AdditionBlock
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.dropout import Dropout
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.layer import LayerError
from pydtnn.model import Model
from pydtnn.tests.abstract.model_common import ModelCommonTestCase
from pydtnn.tests.abstract.common import verbose_test, Params
from pydtnn.utils.tensor import TensorFormat, format_transpose


class ModelTensorTestCase(ModelCommonTestCase):
    """
    Tests that two models with different parameters lead to the same results
    """

    global ModelCommonTestCase

    rtol_dict = ModelCommonTestCase.rtol_dict | {ConcatenationBlock: 1e-0, AdditionBlock: 1e-1, Conv2D: 1e-3}
    atol_dict = ModelCommonTestCase.atol_dict | {ConcatenationBlock: 1e-0, AdditionBlock: 1e-1, Conv2D: 1e-3}

    # NOTE: Delete parent test to prevent re-export and re-testing
    del ModelCommonTestCase

    # Compares results between an XX model {self.model1_desc} and other {self.model1_desc}
    model1_desc = "using the CPU backend tensor format NHWC"
    model2_desc = "using the CPU backend tensor format NCHW"

    def get_model2(self, model_name: str):
        
        # Tensor format NCHW
        params = Params()
        params.model_name = model_name
        params.tensor_format = TensorFormat.NCHW.upper()
        params_dict = vars(params)
        try:
            model2 = Model(**params_dict)
        except LayerError as exc:
            raise unittest.SkipTest(f"Model {model_name} incompatible with {params_dict['dataset_name']}") from exc
        return model2

    @staticmethod
    def nhwc2nchw(x: np.ndarray):
        if len(x.shape) == 4:
            x = format_transpose(x, TensorFormat.NHWC, TensorFormat.NCHW)
        return np.asarray(x, order="C", copy=None)

    def do_model2_forward_pass(self, model2: Model, x1: list[np.ndarray]) -> list[np.ndarray]:
        """
        Model 2 forward pass in NCHW format
        """

        x1_format = list(map(self.nhwc2nchw, x1))
        return super().do_model2_forward_pass(model2, x1_format)

    def do_model2_backward_pass(self, model2: Model, dx1: list[np.ndarray]) -> list[np.ndarray]:
        """
        Model 2 backward pass in NCHW format
        """
        dx1_format = list(map(self.nhwc2nchw, dx1))
        return super().do_model2_backward_pass(model2, dx1_format)

    def compare_forward(self, model1: Model, x1: list[np.ndarray], model2: Model, x2: list[np.ndarray]):
        assert len(x1) == len(x2), "x1 and x2 should have the same length"
        if verbose_test():
            print()
            print(f"Comparing outputs of both models...")
        for i, layer in enumerate(model1.layers, 1):
            # Skip test on layers that behave randomly
            # NOTE: Dropout uses random data and Flatten just reshape the input (it make no sense to undo its work, change the format and flatten again only to compare both formats)
            if not isinstance(layer, (Dropout, Flatten)):
                x1_i = self.nhwc2nchw(x1[i])
                rtol, atol = self.get_tolerance(layer)
                allclose = np.allclose(x1_i, x2[i], rtol=rtol, atol=atol)
                self.assertTrue(allclose,
                                f"Forward result from layers {layer.name_with_id} differ"
                                f" ({self.print_stats(x1_i, x2[i], rtol, atol)})")

    def compare_backward(self, model1: Model, dx1, model2: Model, dx2):
        assert len(dx1) == len(dx2), "dx1 and dx2 should have the same length"
        if verbose_test():
            print()
            print(f"Comparing dw of both models...")
        for i, layer in reversed(list(enumerate(model2.layers, 0))):
            if isinstance(layer, (Conv2D, FC)):
                rtol, atol = self.get_tolerance(layer)
                if len(layer.weights.shape) == 4:
                    # layer.dw: np.ndarray
                    transposed_dw = format_transpose(layer.dw, "OIHW", "IHWO")
                    if transposed_dw.shape == model1.layers[i].dw.shape:
                        allclose = np.allclose(transposed_dw, model1.layers[i].dw, rtol=rtol,
                                               atol=atol)
                        self.assertTrue(allclose,
                                        f"Backward dw from layer {layer.name_with_id} differ"
                                        f" ({self.print_stats(transposed_dw, model1.layers[i].dw, rtol, atol)})")
                else:
                    if layer.dw.shape == model1.layers[i].dw.shape:
                        allclose = np.allclose(layer.dw, model1.layers[i].dw, rtol=rtol, atol=atol)
                        self.assertTrue(allclose,
                                        f"Backward dw from layer {layer.name_with_id} differ"
                                        f" ({self.print_stats(layer.dw, model1.layers[i].dw, rtol, atol)})")
        if verbose_test():
            print()
            print(f"Comparing db of both models...")
        for i, layer in reversed(list(enumerate(model2.layers, 0))):
            if isinstance(layer, (Conv2D, FC)) and layer.use_bias:
                rtol, atol = self.get_tolerance(layer)
                # layer.db:np.ndarray
                allclose = np.allclose(layer.db, model1.layers[i].db, rtol=rtol, atol=atol)
                self.assertTrue(allclose,
                                f"Backward db from layer {layer.name_with_id} differ"
                                f" ({self.print_stats(layer.db, model1.layers[i].db, rtol, atol)})")
        if verbose_test():
            print()
            print(f"Comparing dx of both models...")
        for i, layer in reversed(list(enumerate(model2.layers, 0))):
            # Skip test on layers that behave randomly and Flatten
            if not isinstance(layer, (Dropout, Flatten)):
                rtol, atol = self.get_tolerance(layer)
                dx1_i=self.nhwc2nchw(dx1[i])
                if dx1_i.shape == dx2[i].shape:
                    allclose = np.allclose(dx1_i, dx2[i], rtol=rtol, atol=atol)
                else:
                    warnings.warn(f"dx shape on both models for {layer.name_with_id} differ:"
                                  f" [dx1.shape: {dx1[i].shape}, dx2.shape: {dx2[i].shape}]")
                    # Try flattening both
                    allclose = np.allclose(dx1_i.flatten(), dx2[i].flatten(), rtol=rtol, atol=atol)
                self.assertTrue(allclose,
                                f"Backward result from layer {layer.name_with_id} differ"
                                f" ({self.print_stats(dx1_i, dx2[i], rtol, atol)})")



import numpy as np

from pydtnn.layers.layer_normalization import LayerNormalization
from pydtnn.backends.cpu.layers.layer import LayerCPU


class LayerNormalizationCPU(LayerNormalization[np.ndarray], LayerCPU):
    def forward(self, x):
        mu = np.mean(x, axis=self.axis, keepdims=True)
        xc = (x - mu)
        var = np.mean(xc ** 2, axis=self.axis, keepdims=True)

        self.std = np.sqrt(var + self.epsilon)
        self.xn = xc / self.std
        y = self.gamma * self.xn + self.beta

        return y

    def backward(self, dy):
        self.dgamma = np.sum(dy * self.xn, axis=0)
        self.dbeta = np.sum(dy, axis=0)
        if self.need_dx:
            dy = dy * self.gamma
            dx = dy - self.xn * np.mean(dy * self.xn, self.axis, keepdims=True)
            dx -= np.mean(dy, self.axis, keepdims=True)
            dx /= self.std
            return dx



"""
Performance tests for Conv2D layer using the convGemm library

For running the tests run:
    python profile_conv2d_conv_gemm.py

To obtain a profile, run:
    python3 -m cProfile -o cprofile_conv2d_conv_gemm.prof profile_tests/profile_conv2d_conv_gemm.py

To graphically inspect the profile, run:
    snakeviz cprofile_conv2d_conv_gemm.prof
"""

import cProfile
import time
from copy import deepcopy

import numpy as np

from pydtnn.layers.conv_2d import Conv2D
from pydtnn.model import Model
from pydtnn.utils import random


class D:
    def __init__(self):
        """Default parameters"""
        self.b = 1  # Batch size
        self.c = 1  # Channels per layer
        self.h = 128  # Layers height
        self.w = 100  # Layers width
        self.kn = 1  # Number of filters
        self.kh = 16  # Filters weights height
        self.kw = 10  # Filters weights width
        self.vpadding = 1  # Vertical padding
        self.hpadding = 1  # Horizontal padding
        self.vstride = 1  # Vertical stride
        self.hstride = 1  # Horizontal stride
        self.vdilation = 1  # Vertical dilation
        self.hdilation = 1  # Horizontal dilation

    @property
    def ho(self):
        return (self.h + 2 * self.vpadding - self.vdilation * (self.kh - 1) - 1) // self.vstride + 1

    @property
    def wo(self):
        return (self.w + 2 * self.hpadding - self.hdilation * (self.kw - 1) - 1) // self.hstride + 1

    def __repr__(self):
        return f"""\
x, weights, and y parameters:
  (b, c, h, w)    = {self.b} {self.c} {self.h} {self.w}
  (kn, c, kh, kw) = {self.kn} {self.c} {self.kh} {self.kw}
  (kn, b, ho, wo) = {self.kn} {self.b} {self.ho} {self.wo}
  padding         = {self.vpadding} {self.hpadding}
  stride          = {self.vstride} {self.hstride}
  dilation        = {self.vdilation} {self.hdilation}
"""


def _print_with_header(header, to_be_printed=None):
    print()
    print("-" * (len(header) + 2))
    print(" {}".format(header))
    print("-" * (len(header) + 2))
    if to_be_printed is not None:
        print(to_be_printed)


class Params:
    pass


def get_conv2d_layers(d: D) -> tuple[Conv2D, Conv2D]:
    params = Params()
    params.batch_size = d.b
    params.conv_variant = "i2c"
    model_i2c = Model(**vars(params))
    params_gc = deepcopy(params)
    params_gc.conv_variant = "gemm"
    model_cg = Model(**vars(params_gc))
    conv2d_i2c = Conv2D(nfilters=d.kn, filter_shape=(d.kh, d.kw),
                        padding=(d.vpadding, d.hpadding),
                        stride=(d.vstride, d.hstride), dilation=(d.vdilation, d.hdilation),
                        use_bias=True, weights_initializer="glorot_uniform", biases_initializer="zeros")
    conv2d_i2c.init_backend_from_model(model_i2c)
    conv2d_i2c.debug = True
    conv2d_cg = Conv2D(nfilters=d.kn, filter_shape=(d.kh, d.kw),
                       padding=(d.vpadding, d.hpadding),
                       stride=(d.vstride, d.hstride), dilation=(d.vdilation, d.hdilation),
                       use_bias=True, weights_initializer="glorot_uniform", biases_initializer="zeros")
    conv2d_cg.init_backend_from_model(model_cg)
    conv2d_cg.debug = True
    for layer in (conv2d_i2c, conv2d_cg):
        layer.initialize(prev_shape=(d.c, d.h, d.w))
    # Set the same initial weights and biases to both layers
    conv2d_cg.weights = conv2d_i2c.weights.copy()
    conv2d_cg.biases = conv2d_i2c.biases.copy()
    return conv2d_i2c, conv2d_cg


class PerfTestConv2DConvGemm:
    """
    Performance test for Conv2D that compares Conv2d with mm and i2c.T
    """

    def _test_forward_backward(self, d, x, weights, print_times=False):
        conv2d_i2c, conv2d_cg = get_conv2d_layers(d)
        conv2d_i2c.weights = weights.copy()
        conv2d_cg.weights = weights.copy()

        # Forward i2c
        print("Forward I2C")
        y_i2c = conv2d_i2c.forward(x)  # First run
        profiler = cProfile.Profile()
        profiler.enable()
        y_i2c = conv2d_i2c.forward(x)
        profiler.disable()
        profiler.print_stats()
        profiler.dump_stats("forward_i2c.prof")
        print()

        # Forward cg
        print("Forward CG")
        y_i2c = conv2d_cg.forward(x)  # First run
        profiler = cProfile.Profile()
        profiler.enable()
        y_i2c = conv2d_cg.forward(x)
        profiler.disable()
        profiler.print_stats()
        profiler.dump_stats("forward_cg.prof")
        print()

        dy = random.random((d.b, d.kn, d.ho, d.wo)).astype(np.float32, order='C')

        # Backward i2c
        print("Backward I2C")
        dx_cg = conv2d_i2c.backward(dy)  # First run
        profiler = cProfile.Profile()
        profiler.enable()
        dx_cg = conv2d_i2c.backward(dy)
        profiler.disable()
        profiler.print_stats()
        profiler.dump_stats("backward_i2c.prof")

        # Backward cg with fallback
        print("Backward CG (with fallback)")
        dx_cg = conv2d_cg.backward(dy)  # First run
        profiler = cProfile.Profile()
        profiler.enable()
        dx_cg = conv2d_cg.backward(dy)
        profiler.disable()
        profiler.print_stats()
        profiler.dump_stats("backward_cg_fallback.prof")

        # Backward cg without fallback
        print("Backward CG (without fallback")
        cg_fallback_to_im2col = conv2d_cg.cg_fallback_to_im2col
        conv2d_cg.cg_fallback_to_im2col = False
        dx_cg = conv2d_cg.backward(dy)  # First run
        profiler = cProfile.Profile()
        profiler.enable()
        dx_cg = conv2d_cg.backward(dy)
        profiler.disable()
        profiler.print_stats()
        profiler.dump_stats("backward_cg_no_fallback.prof")
        conv2d_cg.cg_fallback_to_im2col = cg_fallback_to_im2col

        # cg forward
        tic = time.perf_counter()
        y_cg = conv2d_cg.forward(x)
        toc = time.perf_counter()
        print(f"cg forward: {toc - tic:0.4f} s")
        dy = random.random((d.b, d.kn, d.ho, d.wo)).astype(np.float32, order='C')
        # i2c backward
        tic = time.perf_counter()
        dx_i2c = conv2d_i2c.backward(dy)
        toc = time.perf_counter()
        print("---")
        print(f"i2c backward: {toc - tic:0.4f} s")
        print("")
        # cg backward
        tic = time.perf_counter()
        dx_cg = conv2d_cg.backward(dy)
        toc = time.perf_counter()
        print(f"cg backward: {toc - tic:0.4f} s")
        # All close?
        print("dw all close:", np.allclose(conv2d_i2c.dw, conv2d_cg.dw))
        print("dx all close:", np.allclose(dx_i2c, dx_cg))
        print("Y all close:", np.allclose(y_i2c, y_cg, rtol=1e-5, atol=1e-6))

    def test_forward_backward_alexnet_cifar10_first_conv2d(self):
        """Tests that the AlexNet Cifar10 first Conv2d lead to the same solution on i2c and on conv_gemm"""
        d = D()
        d.b = 64
        d.kn, d.kh, d.kw = (64, 3, 3)
        d.c, d.h, d.w = (3, 32, 32)
        d.vpadding, d.hpadding = (1, 1)
        d.vstride, d.hstride = (2, 2)
        d.vdilation, d.hdilation = (1, 1)
        x = random.random((d.b, d.c, d.h, d.w)).astype(np.float32, order='C')
        weights = random.random((d.kn, d.c, d.kh, d.kw)).astype(np.float32, order='C')
        self._test_forward_backward(d, x, weights, print_times=True)

    def test_forward_backward_alexnet_cifar10_second_conv2d(self):
        """Tests that the AlexNet Cifar10 second Conv2d lead to the same solution on i2c and on conv_gemm"""
        d = D()
        d.b = 64
        d.kn, d.kh, d.kw = (192, 3, 3)
        d.c, d.h, d.w = (64, 8, 8)
        d.vpadding, d.hpadding = (1, 1)
        d.vstride, d.hstride = (1, 1)
        d.vdilation, d.hdilation = (1, 1)
        x = random.random((d.b, d.c, d.h, d.w)).astype(np.float32, order='C')
        weights = random.random((d.kn, d.c, d.kh, d.kw)).astype(np.float32, order='C')
        self._test_forward_backward(d, x, weights, print_times=True)

    def test_forward_backward_alexnet_cifar10_third_conv2d(self):
        """Tests that the AlexNet Cifar10 third Conv2d lead to the same solution on i2c and on conv_gemm"""
        d = D()
        d.b = 64
        d.kn, d.kh, d.kw = (192, 3, 3)
        d.c, d.h, d.w = (384, 3, 3)
        d.vpadding, d.hpadding = (1, 1)
        d.vstride, d.hstride = (1, 1)
        d.vdilation, d.hdilation = (1, 1)
        x = random.random((d.b, d.c, d.h, d.w)).astype(np.float32, order='C')
        weights = random.random((d.kn, d.c, d.kh, d.kw)).astype(np.float32, order='C')
        self._test_forward_backward(d, x, weights, print_times=True)

    def test_forward_backward_alexnet_imagenet_first_conv2d(self):
        """Tests that the AlexNet Imagenet first Conv2d lead to the same solution on i2c and on conv_gemm"""
        # id;height;width;channels;kernel_height;kernel_width;kernel_num;stride;padding
        # 2;227;227;3;11;11;96;4;0
        d = D()
        d.b = 64
        d.kn, d.kh, d.kw = (96, 11, 11)
        d.c, d.h, d.w = (3, 227, 227)
        d.vpadding, d.hpadding = (0, 0)
        d.vstride, d.hstride = (4, 4)
        d.vdilation, d.hdilation = (1, 1)
        x = random.random((d.b, d.c, d.h, d.w)).astype(np.float32, order='C')
        weights = random.random((d.kn, d.c, d.kh, d.kw)).astype(np.float32, order='C')
        self._test_forward_backward(d, x, weights, print_times=True)


if __name__ == "__main__":
    c = PerfTestConv2DConvGemm()
    for i in range(1, 2):
        print()
        print("------------")
        print("Iteration:", i)
        print("------------")
        print()
        print("Alexnet for Cifar10")
        print("===================")
        c.test_forward_backward_alexnet_cifar10_first_conv2d()
        # print()
        # print("Alexnet for Imagenet")
        # print("====================")
        # c.test_forward_backward_alexnet_imagenet_first_conv2d()
        # print()
        # print("Alexnet for Cifar10 2n layer")
        # print("============================")
        # c.test_forward_backward_alexnet_cifar10_second_conv2d()
        # print()
        # print("Alexnet for Cifar10 3rd layer")
        # print("=============================")
        # c.test_forward_backward_alexnet_cifar10_third_conv2d()



import numpy as np
from pydtnn.backends.cpu.layers.abstract.conv_2d_standard import Conv2DStandardCPU
from pydtnn.backends.cpu.utils.im2col_nchw_cython import col2im_nchw_cython, im2col_nchw_cython#, alt_col2im_nchw_cython
from pydtnn.backends.cpu.utils.im2row_nhwc_cython import im2row_nhwc_cython, row2im_nhwc_cython#, alt_row2im_nhwc_cython

from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum

from pydtnn.utils.constants import ArrayShape
from pydtnn.utils.tensor import TensorFormat, format_transpose


class Conv2DI2CCPU(Conv2DStandardCPU):

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None) -> None:
        super().initialize(prev_shape, x)

        # self.dim_n: Dimension where the "n" of NCHW/NHWC is used in the calculations.
        # self.dim_c: Dimension where the "c" of NCHW/NHWC is used in the calculations.
        self.dim_n = self.model.batch_size * self.ho * self.wo
        self.dim_c = self.ci * self.kh * self.kw

        match self.model.tensor_format:
            case TensorFormat.NCHW:
                self.forward = self._forward_i2c_nchw
                self.backward = self._backward_i2c_nchw
                self._x_cols = np.zeros(shape=(self.dim_c, self.dim_n), dtype=self.model.dtype, order="C")

                _dw_shape = (self.co, self.dim_c)
                res_bw_shape = (self.dim_c, self.dim_n)
                dx_shape = (self.model.batch_size, self.ci, self.hi, self.wi)
            case TensorFormat.NHWC:
                self.forward = self._forward_i2c_nhwc
                self.backward = self._backward_i2c_nhwc
                self._x_rows = np.zeros(shape=(self.dim_n, self.dim_c), dtype=self.model.dtype, order="C")                

                _dw_shape = (self.dim_c, self.co)
                res_bw_shape = (self.dim_n, self.dim_c)
                dx_shape = (self.model.batch_size, self.hi, self.wi, self.ci)
            case _:
                _dw_shape = (None, )
                res_bw_shape = (None, )
                dx_shape = (None,)

                raise NotImplementedError(f"\"{self.model.tensor_format}\" format not implemented.")
        # -

        # NOTE: These attributes only store data, their values before the operation doesn't matter; they're initalized due avoid warnings in "LayerAndActivationBase.export".
        self.dx = np.zeros(shape=dx_shape, dtype=self.model.dtype, order="C")
        self.res = np.zeros(shape=(self.dim_n, self.co), dtype=self.model.dtype, order="C")
        self._dw = np.zeros(shape=_dw_shape, dtype=self.model.dtype, order="C")
        self.res_bw = np.zeros(shape=res_bw_shape, dtype=self.model.dtype, order="C")
    # ---

    def _forward_i2c_nhwc(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses im2col and matmul"""

        dim_n = x.shape[0] * self.ho * self.wo
        # x_rows = np.zeros(shape=(dim_n, self.dim_c), dtype=self.model.dtype)
        x_rows = np.asarray(self._x_rows[:dim_n, :], dtype=self.model.dtype, order="C", copy=None)
        res = self.res[:dim_n, :]

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_IM2COL)
        im2row_nhwc_cython(x, x_rows,
                           self.kh, self.kw, self.ho, self.wo,
                           self.vpadding, self.hpadding,
                           self.vstride, self.hstride, self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.x_rows = x_rows

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_RESHAPE_W)
        w_cols = self.weights.reshape((-1, self.co), copy=False)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_MATMUL)
        np.matmul(x_rows, w_cols, out=res,
                  dtype=self.model.dtype)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_SUM_BIASES)
            np.add(res, self.biases.reshape((-1, self.co), copy=False), out=res,
                   dtype=self.model.dtype)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_RESHAPE_Y)
        y = res.reshape((-1, self.ho, self.wo, self.co), copy=False)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def _forward_i2c_nchw(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses im2col and matmul"""

        dim_n = x.shape[0] * self.ho * self.wo
        # x_cols = np.zeros(shape=(self.dim_c, dim_n), dtype=self.model.dtype)
        x_cols = np.asarray(self._x_cols[:, :dim_n], dtype=self.model.dtype, order="C", copy=None)
        res = self.res[:dim_n, :]

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_IM2COL)
        im2col_nchw_cython(x, x_cols,
                           self.kh, self.kw, self.ho, self.wo,
                           self.vpadding, self.hpadding,
                           self.vstride, self.hstride, self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.x_cols = x_cols

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_RESHAPE_W)
        w_rows = self.weights.reshape((self.co, -1), copy=False)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_MATMUL)
        np.matmul(w_rows, x_cols, out=res.T,
                  dtype=self.model.dtype)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_SUM_BIASES)
            np.add(res, self.biases.reshape((-1, self.co), copy=False), out=res,
                   dtype=self.model.dtype)

            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_RESHAPE_Y)
        y: np.ndarray = format_transpose(res.reshape((-1, self.ho, self.wo, self.co), copy=False), "NHWC", "NCHW")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def _backward_i2c_nhwc(self, dy: np.ndarray) -> np.ndarray:
        """Version of the backward function that uses im2col and matmul"""

        res = np.asarray(self.res_bw[:(dy.shape[0] * self.ho * self.wo), :], dtype=self.model.dtype, order="C", copy=None)
        
        dx = self.dx[:dy.shape[0], :]
        dx.fill(0)  # NOTE: It is necessary that dx is filled with 0s.

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_TRANSPOSE_DY)
        dy_cols: np.ndarray = dy.reshape((-1, self.co), copy=False)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        # Weigths gradient
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DW_MATMUL)
        np.matmul(self.x_rows.T, dy_cols, out=self._dw,
                  dtype=self.model.dtype)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_RESHAPE_DW)
        self.dw = self._dw.reshape(self.weights.shape, copy=False, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        # Biases gradient
        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_SUM_BIASES)
            np.sum(dy, axis=(0, 1, 2), out=self.db)
            # np.sum(dy.reshape((self.co, -1), copy=None), axis=1, out=self.db)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        # Data gradient
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_TRANSPOSE_W)
        w_rows = self.weights.reshape((-1, self.co), copy=False).T
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DX_MATMUL)
        np.matmul(dy_cols, w_rows, out=res,
                  dtype=self.model.dtype)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DX_COL2IM)
        row2im_nhwc_cython(res, dx,
                           dy.shape[0], self.hi, self.wi, self.ci,
                           self.kh, self.kw, self.ho, self.wo,
                           self.vpadding, self.hpadding,
                           self.vstride, self.hstride, self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(dx, dtype=self.model.dtype, order='C', copy=None)

    def _backward_i2c_nchw(self, dy: np.ndarray) -> np.ndarray:
        """Version of the backward function that uses im2col and matmul"""
        res = np.asarray(self.res_bw[:, :(dy.shape[0] * self.ho * self.wo)], dtype=self.model.dtype, order="C", copy=None)

        dx = self.dx[:dy.shape[0], :]
        dx.fill(0)  # NOTE: It is necessary that dx is filled with 0s.

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_TRANSPOSE_DY)
        dy_rows: np.ndarray = format_transpose(dy, "NCHW", "CNHW").reshape((self.co, -1), copy=None)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        # Weigths gradient
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DW_MATMUL)
        np.matmul(dy_rows, self.x_cols.T, out=self._dw,
                  dtype=self.model.dtype)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_RESHAPE_DW)
        self.dw = self._dw.reshape(self.weights.shape, copy=True, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        # Biases gradient
        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_SUM_BIASES)
            np.sum(dy, axis=(0, 2, 3), out=self.db)
            # np.sum(dy.reshape((self.co, -1), copy=False), axis=1, out=self.db)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        # Data gradient
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_TRANSPOSE_W)
        w_cols = self.weights.reshape((self.co, -1), copy=False).T
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DX_MATMUL)
        np.matmul(w_cols, dy_rows, out=res,
                  dtype=self.model.dtype, order='C')
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DX_COL2IM)
        col2im_nchw_cython(res, dx,
                           dy.shape[0], self.ci, self.hi, self.wi,
                           self.kh, self.kw, self.ho, self.wo,
                           self.vpadding, self.hpadding,
                           self.vstride, self.hstride, self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(dx, dtype=self.model.dtype, order='C', copy=None)



from pydtnn.layers.abstract.block_layer import AbstractBlockLayer
from pydtnn.utils.constants import Array, ArrayShape


class FeedForward[T: Array](AbstractBlockLayer[T]):
    def __init__(self, shape: ArrayShape = (1,), d_ff: int = 256, dropout_rate: float = 0.5):
        super().__init__()
        self.shape = shape
        self.d_ff = d_ff
        self.dropout_rate = dropout_rate



import numpy as np

from pydtnn.backends.cpu.metrics.binary_confusion_matrix import BinaryConfusionMatrixCPU
from pydtnn.backends.cpu.metrics.metric import MetricCPU
from pydtnn.metrics.precision import Precision
from pydtnn.backends.cpu.utils.div_arrays_set_if_zero import div_arrays_set_if_zero


class PrecisionCPU(Precision[np.ndarray], MetricCPU):

    conf_matrix_metric: BinaryConfusionMatrixCPU

    def compute(self, y_pred: np.ndarray, y_targ: np.ndarray) -> float:
        true_positives = self.conf_matrix_metric.get_true_positives()
        false_positives = self.conf_matrix_metric.get_false_positives()
        # true_positives / (true_positives + false_positives)
        precision = np.asarray(true_positives, dtype=np.dtype(float), order="C", copy=True)
        divider = np.add(true_positives, false_positives, dtype=np.dtype(float), order="C")
        div_arrays_set_if_zero(precision,  divider, default_value=0.0)
        return float(np.average(precision))


from pydtnn.activations.activation import Activation
from pydtnn.utils.constants import Array, ArrayShape


class Softmax[T: Array](Activation[T]):
    def __init__(self, shape: ArrayShape = (1,), axis: int = 1):
        super().__init__(shape)
        self.axis_dim = axis



from typing import Callable, List

from pydtnn.backends.cpu.layers.conv_2d.direct_cpu import Conv2DDirectCPU
from pydtnn.backends.cpu.layers.conv_2d.winograd_cpu import Conv2DWinogradCPU
from pydtnn.model import Model
from pydtnn.utils.best_of import BestOf

import numpy as np
from pydtnn.utils.constants import ArrayShape



# FIXME: Broken since Conv2D to backend support
class BestOfVariant(Conv2DWinogradCPU, Conv2DDirectCPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # best_of related attributes (will be initialized in initialize())
        self._best_fw: BestOf = None # type: ignore
        self._best_fw_bw_pipeline: BestOf = None # type: ignore
        # Other parameters
        self.variant = None

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None):
        super().initialize(prev_shape, x)
        if self.model.enable_best_of:
            # Set variant to 'best_of' and set alternatives to only forward, and forward backward best_ofs
            self.variant = Conv2DDirectCPU.Variant.BEST_OF
            # Bestof will honor the next configuration options:
            # - enable_conv_winograd
            # - enable_conv_gemm
            # - enable_conv_direct
            # - conv_direct_methods_for_best_of (if empty, conv_direct_method will be used instead)
            # Set alternatives for only forward, and for forward backward
            alternatives_fw = []
            alternatives_fw_bw_pipeline = []
            if self.model.enable_conv_i2c:
                alternatives_fw.append((Conv2DDirectCPU.Variant.I2C, self._get_class_forward_and_backward(Conv2DDirectCPU.Variant.I2C)[0]))
                alternatives_fw_bw_pipeline.append((Conv2DDirectCPU.Variant.I2C, self._get_class_forward_and_backward(Conv2DDirectCPU.Variant.I2C)))
            if self.model.enable_conv_gemm:
                alternatives_fw.append((Conv2DDirectCPU.Variant.GEMM, self._get_class_forward_and_backward(Conv2DDirectCPU.Variant.GEMM)[0]))
                alternatives_fw_bw_pipeline.append((Conv2DDirectCPU.Variant.GEMM, self._get_class_forward_and_backward(Conv2DDirectCPU.Variant.GEMM)))
            if self.model.enable_conv_winograd and self.cw_constraints_fulfilled:
                alternatives_fw.append((Conv2DDirectCPU.Variant.WINOGRAD, self._get_class_forward_and_backward(Conv2DDirectCPU.Variant.WINOGRAD)[0]))
                alternatives_fw_bw_pipeline.append((Conv2DDirectCPU.Variant.WINOGRAD, self._get_class_forward_and_backward(Conv2DDirectCPU.Variant.WINOGRAD)))
            if self.model.enable_conv_direct:
                for n in range(len(self.cd)):
                    cdn = f"cd{n}"
                    alternatives_fw.append((cdn, self._get_class_forward_and_backward(cdn)[0]))
                    alternatives_fw_bw_pipeline.append((cdn, self._get_class_forward_and_backward(cdn)))
            self._best_fw = BestOf(
                name="Conv2DCPU only forward",
                alternatives=alternatives_fw,
                get_problem_size=lambda *args: tuple(args[0].shape) + tuple(args[0].weights.shape)
                + (args[0].vstride,
                   args[0].hstride,
                   args[0].vdilation,
                   args[0].hdilation),
            )
            self._best_fw_bw_pipeline = BestOf(
                name="Conv2DCPU forward backward",
                alternatives=alternatives_fw_bw_pipeline,
                get_problem_size=lambda *args: tuple(args[0].shape) + tuple(args[0].weights.shape)
                + (args[0].vpadding, args[0].hpadding,
                   args[0].vstride, args[0].hstride,
                   args[0].vdilation, args[0].hdilation),
            )

    def _get_class_forward_and_backward(self, variant) -> List[Callable]:
        return [getattr(self.__class__, f'_forward_{variant}_{self.model.tensor_format}'),
                getattr(self.__class__, f'_backward_{variant}_{self.model.tensor_format}')]

    def _fw_bw_best_of(self, stage, x_or_y):
        match self.model.mode:
            case Model.Mode.TRAIN:
                return self._best_fw_bw_pipeline(stage, self, x_or_y)
            case Model.Mode.EVALUATE:
                return self._best_fw(self, x_or_y)
            case _:
                raise RuntimeError("Conv2D BestOf variant requires Model.mode to be set to ModelModeEnum.EVALUATE or ModelModeEnum.TRAIN")

    def _forward_best_of_nhwc(self, x):
        return self._fw_bw_best_of(0, x)

    def _forward_best_of_nchw(self, x):
        return self._fw_bw_best_of(0, x)

    def _backward_best_of_nhwc(self, y):
        return self._fw_bw_best_of(1, y)

    def _backward_best_of_nchw(self, y):
        return self._fw_bw_best_of(1, y)



"""Fused layers test group"""

from pydtnn.tests.conv_2d_relu import Conv2DReluTestCase
from pydtnn.tests.conv_2d_batch_normalization import Conv2DBatchNormalizationTestCase
from pydtnn.tests.conv_2d_batch_normalization_relu import Conv2DBatchNormalizationReluTestCase
from pydtnn.tests.batch_normalization_relu import BatchNormalizationReluTestCase



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Given an output file from pydtnn_benchmark.py with the profile option
activated, prints a csv file with the profile information.
"""

###########################################################################
# IMPORTS                                                                 #
###########################################################################
import re
import sys
import getopt
import pathlib
import gzip
import tempfile


###########################################################################
# MISCELLANEOUS FUNCTIONS                                                 #
###########################################################################
def my_help():
    """Print the the command line usage help."""
    print("""Usage: extract_profile_from_output.py [OPTION]... FILE

Given an output file from pydtnn_benchmark.py with the profile option
activated, prints a csv file with the profile information.

Options:
    -v, --VERBOSE      increment the output verbosity.
    -h, --help         display this help and exit.

Please, report bugs to <barrachi@uji.es>.
""")


def log(text):
    """Log a message to stderr."""
    print(f">>> {text}\n", file=sys.stderr)


def error(text):
    """Report an error message and exit."""
    print(f"ERROR: {text}\n", file=sys.stderr)
    raise SystemExit(-1)


# Global command line parameters
INPUT_FILE_NAME = None
VERBOSE = 0
SCRIPT_PATH = pathlib.Path(__file__).parent.absolute()


def get_opts():
    """Read command line options."""
    global INPUT_FILE_NAME, VERBOSE
    optlist, args = getopt.getopt(sys.argv[1:],
                                  'hv',
                                  ['VERBOSE', 'help'])
    for opt, arg in optlist:
        if opt in ('-h', '--help'):
            my_help()
            raise SystemExit()
        elif opt in ('-v', '--verbosity'):
            VERBOSE = 1
    # Check required arguments
    if len(args) == 0:
        my_help()
        error("At least a FILE is required")
    INPUT_FILE_NAME = args[0]


###########################################################################
# APPLICATION SPECIFIC FUNCTIONS                                          #
###########################################################################
def print_profile_from_file(file):
    """Do print the profile part from file."""
    first_line_re = re.compile("ncalls *tottime")
    in_profile_section = False
    for line in file.readlines():
        if first_line_re.search(line):
            in_profile_section = True
        if in_profile_section:
            line = line.strip()
            if line == "":
                return
            line = line.split(None, 5)
            print(",".join(line))


def print_profile():
    """Get the file and print the profile"""
    if INPUT_FILE_NAME is None:
        with sys.stdin as file:
            print_profile_from_file(file)
    else:
        file_open = open if INPUT_FILE_NAME[-3:] != ".gz" else gzip.open
        with file_open(INPUT_FILE_NAME, 'rt') as file:
            print_profile_from_file(file)


def main():
    """Do the work (main function, called when not imported)."""
    get_opts()
    # Main part of the application
    print_profile()


if __name__ == "__main__":
    main()



import numpy as np

from pydtnn.backends.cpu.losses.loss import LossCPU
from pydtnn.losses.binary_cross_entropy import BinaryCrossEntropy


class BinaryCrossEntropyCPU(BinaryCrossEntropy[np.ndarray], LossCPU):
    def compute(self, _y_pred: np.ndarray, y_targ: np.ndarray, batch_size: int) -> tuple[float, np.ndarray]:
        assert len(y_targ.shape) == 2
        # Loss
        b = y_targ.shape[0]
        loss: float = -np.sum(np.log(np.maximum((1 - y_targ) - _y_pred, self.eps))) / b

        # Dx
        y_pred: np.ndarray = np.clip(_y_pred, a_min=self.eps, a_max=(1 - self.eps))
        dx: np.ndarray = (-(y_targ / y_pred) + ((1 - y_targ) / (1 - y_pred))) / batch_size
        return loss, np.asarray(dx, dtype=self.model.dtype, order="C", copy=None)



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]

# TODO: Missing `bn_inference_nchw_cython`


def bn_inference_cython[T: _npDT](x: _npDT_2Dims[T],
                                  y: _npDT_2Dims[T],
                                  running_mean: _npDT_1Dims[T],
                                  std: _npDT_1Dims[T],
                                  gamma: _npDT_1Dims[T],
                                  beta: _npDT_1Dims[T]) -> None:
    """
    Args:
        x (npDT_2Dims): The 2 dimensional input's ndarray.
        y (npDT_2Dims): The 2 dimensional outputs's ndarray.
        running_mean (npDT_1Dims): The 1 dimensions ndarray that stores the running mean.
        std (npDT_1Dims): The input's 1 dimensions thtat stores the standard deviation
        gamma (npDT_1Dims): The input's 1 dimensions ndarray the gamma's values
        beta (npDT_1Dims): The input's 1 dimensions ndarray the beta's values

    Returns:
        Nothing. The output is stored in "y".
    """


def bn_relu_inference_cython[T: _npDT](x: _npDT_2Dims[T],
                                       y: _npDT_2Dims[T],
                                       running_mean: _npDT_1Dims[T],
                                       inv_std: _npDT_1Dims[T],
                                       gamma: _npDT_1Dims[T],
                                       beta: _npDT_1Dims[T]) -> None:
    """
    Args:
        x (npDT_2Dims): The 2 dimensional input's ndarray.
        y (npDT_2Dims): The 2 dimensional output's ndarray.
        running_mean (npDT_1Dims): The 1 dimensions ndarray that stores the running mean.
        inv_std (npDT_1Dims): The input's 1 dimensions thtat stores the inverse standard deviation
        gamma (npDT_1Dims): The input's 1 dimensions ndarray the gamma's values
        beta (npDT_1Dims): The input's 1 dimensions ndarray the beta's values

    Returns:
        Nothing. The output will be stored in \"y\".
    """



"""
PyDTNN parser

The parser in this module will be used by 'pydtnn_benchmark' to parse the
command line arguments.

And what is even more important, it will also be loaded by the Model class to
obtain default values to its non-mandatory attributes. This way, when a model
object is instantiated (even if it is not from 'pydtnn_benchmark') it will
initially have default values for all the attributes declared on the self.

If you want to define a new option, just declare it here. It will automatically
be available as a Model attribute.
"""

import argparse
import multiprocessing
import os

import numpy as np
from pydtnn.utils import parse_bool as bool_lambda
from functools import cache

from typing import Any


def factor(x):
    """Returns x, which must be 0.0 < x <= 1.0"""
    x = float(x)
    if not (0.0 < x <= 1.0):
        raise ValueError("Provided value must be greater than 0.0 and less or equal to 1.0")
    return x


def np_dtype(x):
    """Returns a numpy object from a string representing the data type"""
    return getattr(np, x)


_this_file_path = os.path.dirname(os.path.realpath(__file__))
_scripts_path = os.path.join(_this_file_path, "scripts")
_default_dataset_path = os.path.join(_this_file_path, "datasets/mnist")
_desc = "Trains or evaluates a neural network using PyDTNN."
_epilogue = f"""Example scripts that call this program for training
and evaluating different neural network models with different datasets are
available at: '{_scripts_path}'."""


def _get_mpi_processes():
    try:
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        from pympi import MPI
    except Exception:
        mpi_processes = 1
    else:
        mpi_processes = MPI.COMM_WORLD.Get_size()
    return mpi_processes


def _get_threads_per_process():
    #  From IBM OpenMP documentation: If you do not set OMP_NUM_THREADS, the number of processors available is the
    #  default value to form a new team for the first encountered parallel construct.
    threads_per_process = os.environ.get("OMP_NUM_THREADS", multiprocessing.cpu_count())
    return threads_per_process


def _get_gpus_per_node():
    import subprocess
    try:
        gpus_per_node = subprocess.check_output(["nvidia-smi", "-L"]).count(b'UUID')
    except (FileNotFoundError, subprocess.CalledProcessError):
        gpus_per_node = 0
    return gpus_per_node


def _get_mpi_protocol():
    try:
        from pydtnn.libs.libmpi import proto as PROTOCOL
        from pydtnn.libs.libmpi import ssl as SSL
    except Exception as e:
        PROTOCOL = None
        SSL = None
    protocol = str(PROTOCOL)
    if PROTOCOL and SSL:
        protocol = f"{protocol}+tls"
    return protocol


def _get_mpi_server():
    try:
        from pydtnn.libs.libmpi import addr
    except Exception as e:
        addr = None
    return addr


def _get_mpi_port():
    try:
        from pydtnn.libs.libmpi import port
    except Exception as e:
        port = None
    return port

# NOTE: with @cache it's not possible to extend the class.


@cache  # <== Singleton
class PydtnnArgumentParser(argparse.ArgumentParser):
    lines = []

    def __init__(self):
        super().__init__(description=_desc, epilog=_epilogue)
        # Parser and the supported arguments with their default values
        # (argparse.SUPPRESS is used to avoid showing them on the message)

        # Model
        self.add_argument('--model', dest="model_name", type=str, default=None,
                          help="Neural network model: \'simplemlp\', \'simplecnn\', \'alexnet\', \'vgg11\', \'vgg16\', etc. Default: \'None\'.")
        self.add_argument('--batch-size', type=int, default=None,
                          help="Batch size per MPI rank. Or \'batch_size\' or \'global_batch_size\' must have a value different from \'None\' (but not both). Default: \'None\'.")
        self.add_argument('--global-batch-size', type=int, default=None,
                          help="Batch size between all MPI ranks. Or \'batch_size\' or \'global_batch_size\' must have a value different from \'None\' (but not both) Default: \'None\'.")
        self.add_argument('--dtype', type=np_dtype, default=np.float32,
                          help="Datatype to use: \'float32\', \'float64\'. Default: float32.")
        self.add_argument('--num-epochs', type=int, default=1,
                          help="Number of epochs to perform. Default: 1.")
        self.add_argument('--steps-per-epoch', type=float, default=0,
                          help="Trims the training data depending on the given number of steps per epoch. Default: 0, i.e., do not trim.")
        self.add_argument('--evaluate', dest="evaluate_on_train", default=False, type=bool_lambda,
                          help="Evaluate the model before and after training the model. Default: False.")
        self.add_argument('--evaluate-only', default=False, type=bool_lambda,
                          help="Only evaluate the model. Default: False.")
        self.add_argument('--weights-and-bias-filename', type=str, default=None,
                          help="Load weights and bias from file. Default: None.")
        self.add_argument('--history-file', type=str, default=None,
                          help="Filename to save training loss and metrics.")
        self.add_argument('--tensor-format', type=lambda s: str(s).upper(), default="NHWC",
                          help="Data format to be used: \'NHWC\' or \'NCHW\'. Optionally, the \'AUTO\' value sets \'NCHW\' when the option \'--enable-gpu\' is set and \'NHWC\' otherwise. Default: \'NHWC\'.")
        self.add_argument('--random-seed', type=int, default=57005,
                          help='Initial state of random number generator. Default: 57005.')

        # Synchronization options
        _sy_group = self.add_argument_group("Synchronization options")
        _sy_group.add_argument('--shared-storage', default=True, type=bool_lambda,
                               help="If \'True\' ranks assume they share the file system. Default: True.")
        _sy_group.add_argument('--model-sync-freq', type=int, default=0,
                               help="Number of batches between model syncronization. The \'0\' value syncronizes gradients every batch. Positive values syncronizes gradients and weights every N batches. Negative values disables syncronization. Default: 0.")
        _sy_group.add_argument('--model-sync-alg', type=str, default="avg", choices=["avg", "wavg", "invwavg"],
                               help="Aggregation method used to syncronize models: \'avg\', \'wavg\' or \'invwavg\'. Default: \'avg\'.")
        _sy_group.add_argument('--model-sync-participation', type=str, default="all", choices=["all", "avail2all"],
                               help="Rank participation to syncronize models: \'all\' or \'avail2all\'. Default: \'all\'.")
        _sy_group.add_argument('--model-sync-min-avail', type=int, default=0,
                               help="Minumun ranks with data required to syncronize models. Default: 0.")
        _sy_group.add_argument('--initial-model-sync', type=bool_lambda, default=True,
                               help="Sincronize models on training start. Default: True.")
        _sy_group.add_argument('--final-model-sync', type=bool_lambda, default=True,
                               help="Sincronize models on training end. Default: True.")

        # Dataset options
        _ds_group = self.add_argument_group("Dataset options")
        _ds_group.add_argument('--dataset', dest="dataset_name", type=str, default=None, choices=["mnist", "cifar10", "imagenet", "archive", "folder", "chestxray", "synthetic"],
                               help="Dataset to train: \'mnist\', \'cifar10\', \'cyclone\', \'tsunamis\', \'imagenet\', \'archive\', \'folder\' \'chestxray\' or \'synthetic\'. Default: \'None\'.")
        _ds_group.add_argument('--dataset-percentage', type=float, default=0.0,
                               help="Percentage of dataset that will be used. If it is \'0\': it is deactivated; if is is a value below \'1\' (and above 0): it will perform undersampling; and if is is a value above \'1\': it will perform oversampling. Default: 0.")
        _ds_group.add_argument('--dataset-path', type=str, default=_default_dataset_path,
                               help="Path to the dataset.")
        _ds_group.add_argument('--dataset-lang', type=str, default="en",
                               help="Dataset language.")
        _ds_group.add_argument('--dataset-lang2', type=str, default="de",
                               help="Dataset second language.")
        _ds_group.add_argument('--synthetic-train-samples', default=1000, type=int,
                               help="Number of synthetic train sample. Default: 1000.")
        _ds_group.add_argument('--synthetic-test-samples', default=100, type=int,
                               help="Number of synthetic train sample. Default: 100.")
        _ds_group.add_argument('--synthetic-input-shape', default="3,32,32", type=str,
                               help="Number of synthetic input shape (coma separated). Default: 3,32,32.")
        _ds_group.add_argument('--synthetic-output-shape', default="10", type=str,
                               help="Number of synthetic input shape (coma separated). Default: 10.")
        _ds_group.add_argument('--test-as-validation', default=False, type=bool_lambda,
                               help="Prevent making partitions on training data for training+validation data, use test data for validation. True if specified.")
        _ds_group.add_argument('--validation-split', type=factor, default=0.2,
                               help="Split between training and validation data.")
        _ds_group.add_argument('--augment-shuffle', default=True, type=bool_lambda,
                               help="Shuffle training images. Default: True.")
        _ds_group.add_argument('--augment-flip', default=False, type=bool_lambda,
                               help="Flip horizontally training images. Default: False.")
        _ds_group.add_argument('--augment-flip-prob', type=factor, default=0.5,
                               help="Probability to flip training images. Default: 0.5.")
        _ds_group.add_argument('--augment-crop', default=False, type=bool_lambda,
                               help="Crop training images. Default: False.")
        _ds_group.add_argument('--augment-crop-size', type=int, default=16,
                               help="Size to crop training images. Default: 16.")
        _ds_group.add_argument('--augment-crop-prob', type=factor, default=0.5,
                               help="Probability to crop training images. Default: 0.5.")
        _ds_group.add_argument('--transform-crop', default=False, type=bool_lambda,
                               help="Crop the images. True if specified.")
        _ds_group.add_argument('--transform-crop-perc', type=float, default=0.875,
                               help="Central crop of the images. Default: 0.875.")
        _ds_group.add_argument('--transform-resize', default=False, type=bool_lambda,
                               help="Resize the images. True if specified.")
        _ds_group.add_argument('--transform-resize-size', type=int, default=300,
                               help="New size of the images. Default: 300.")
        _ds_group.add_argument('--normalize', default=False, type=bool_lambda,
                               help="Normalize dataset. Default: False.")
        _ds_group.add_argument('--normalize-offset', type=float, default=-0.45,
                               help="Offset samples by a value. Default: -0.45.")
        _ds_group.add_argument('--normalize-scale', type=float, default=3.75,
                               help="Scale samples by a value. Default: 3.75.")

        # Optimization options
        _oo_group = self.add_argument_group("Optimization options")
        _oo_group.add_argument('--enable-best-of', type=bool_lambda, default=False,
                               help="Enable the BestOf auto-tuner.")
        _oo_group.add_argument('--enable-memory-cache', type=bool_lambda, default=True,
                               help="Enable the memory cache module to use persistent memory.")
        _oo_group.add_argument('--enable-fused-bn-relu', type=bool_lambda, default=False,
                               help="Fuse BatchNormalization and Relu layers. True if specified.")
        _oo_group.add_argument('--enable-fused-conv-relu', type=bool_lambda, default=False,
                               help="Fuse Conv2D and Relu layers. True if specified.")
        _oo_group.add_argument('--enable-fused-conv-bn', type=bool_lambda, default=False,
                               help="Fuse Conv2D and BatchNormalization layers. True if specified.")
        _oo_group.add_argument('--enable-fused-conv-bn-relu', type=bool_lambda, default=False,
                               help="Fuse Conv2D and BatchNormalization and Relu layers. Default: False.")

        # Convolution methods
        _cm_group = self.add_argument_group("Convolution options")
        _cm_group.add_argument('--conv-variant', type=str, default="i2c", choices=["i2c", "gemm", "winograd", "direct"],
                               help="Select the standard 2D Convolutional module. Options: \n" \
                                    "* \'i2c\': Use the ConvI2C algorithm. \n" \
                                    "* \'gemm\': Use the ConvGemm algorithm. \n" \
                                    "* \'winograd\': Use the CondWinograd algorithm. \n" \
                                    "* \'direct\': Use the ConvDirect algorithm. \n" \
                                    "Default: \"i2x\".")
        _cm_group.add_argument('--conv-direct-method', type=str, default="",
                               help="Use ConvDirect module to realize convolutions in Conv2D layers. True if specified.")
        _cm_group.add_argument('--conv-direct-methods-for-best-of', type=str, default="",
                               help="ConvDirect modules to compare in \'best_of\' option if specified.")

        # Optimizer options
        _op_group = self.add_argument_group("Optimizer options")
        _op_group.add_argument('--optimizer', dest="optimizer_name", type=str, default="sgd", choices=["sgd", "rmsprop", "adam", "nadam", "oktopk"],
                               help="Optimizers: \'sgd\', \'rmsprop\', \'adam\', \'nadam\', \'oktopk\'. Default: \'sgd\'. ")
        _op_group.add_argument('--learning-rate', type=float, default=1e-2,
                               help="Learning rate. Default: 0.01.")
        _op_group.add_argument('--learning-rate-scaling', default=False, type=bool_lambda,
                               help="Scale learning rate in data parallelism: new_lr = lr / num_procs.  True if specified.")
        _op_group.add_argument('--optimizer-momentum', type=float, default=0.9,
                               help="Decay rate for \'sgd\' optimizer. Default: 0.9. optimizers. Default: 1e-8.")
        _op_group.add_argument('--optimizer-decay', type=float, default=0.0,
                               help="Decay rate for optimizers. Default: 0.0.")
        _op_group.add_argument('--optimizer-nesterov', default=False, type=bool_lambda,
                               help="Whether to apply Nesterov momentum. Default: False.")
        _op_group.add_argument('--optimizer-beta1', type=float, default=0.99,
                               help="Variable for \'adam\', \'nadam\' optimizers. Default: 0.99.")
        _op_group.add_argument('--optimizer-beta2', type=float, default=0.999,
                               help="Variable for \'adam\', \'nadam\' optimizers. Default: 0.999.")
        _op_group.add_argument('--optimizer-epsilon', type=float, default=1e-7,
                               help="Variable for \'rmsprop\', \'adam\', \'nadam\'. Default=1e-7.")
        _op_group.add_argument('--optimizer-rho', type=float, default=0.9,
                               help="Variable for \'rmsprop\' optimizers. Default: 0.99.")
        _op_group.add_argument('--optimizer-tau', type=int, default=64,
                               help="Variable for \'oktopk\' optimizers. Default: 64.")
        _op_group.add_argument('--optimizer-tau-prime', type=int, default=32,
                               help="Variable for \'oktopk\' optimizers. Default: 32.")
        _op_group.add_argument('--optimizer-density', type=float, default=0.01,
                               help="Variable for \'oktopk\' optimizers. Default: 0.01.")
        _op_group.add_argument('--loss-func', dest="loss_func_name", type=str, default="categorical_cross_entropy",
                               choices=["categorical_cross_entropy", "binary_cross_entropy"],
                               help="Loss functions that is evaluated on each trained batch: \'categorical_cross_entropy\', \'binary_cross_entropy\'. Default \'categorical_cross_entropy\'.")
        _op_group.add_argument(
            '--metrics',
            type=str,
            default="categorical_accuracy",
            help="List of comma-separated metrics that are evaluated on each trained batch: \'categorical_accuracy\', \'categorical_hinge\', \'categorical_mse\', \'categorical_mae\', \'regression_mse\', \'regression_mae\', \'binary_confusion_matrix\', \'multiclass_confusion_matrix\', \'precision\', \'recall\', \'f1_score\'. Default: \'categorical_accuracy\'.")

        # Schedulers options
        _sh_group = self.add_argument_group("Schedulers options")
        _sh_group.add_argument('--schedulers', dest="schedulers_names", type=str,
                               default="early_stopping,reduce_lr_on_plateau,model_checkpoint",
                               help="List of comma-separated LR schedulers: \'warm_up\', \'early_stopping\', \'reduce_lr_on_plateau\', \'reduce_lr_every_nepochs\', \'model_checkpoint\'. Default: \'early_stopping,reduce_lr_on_plateau,model_checkpoint\'.")
        _sh_group.add_argument('--warm-up-epochs', type=int, default=5,
                               help="Number of batches (ramp up) that the LR is scaled up from 0 until LR. Default: 5.")
        _sh_group.add_argument('--early-stopping-metric', type=str, default="val_categorical_cross_entropy",
                               help="Loss metric monitored by early_stopping LR scheduler. Default: \'val_categorical_cross_entropy\'.")
        _sh_group.add_argument('--early-stopping-patience', type=int, default=10,
                               help="Number of epochs with no improvement after which training will be stopped. Default: 10.")
        _sh_group.add_argument('--early-stopping-minimize', type=bool_lambda, default=True,
                               help="Whether to minize the metric. If False, it will maximize. Default: True.")
        _sh_group.add_argument('--reduce-lr-on-plateau-metric', type=str, default="val_categorical_cross_entropy",
                               help="Loss metric monitored by reduce_lr_on_plateau LR scheduler. Default: \'val_categorical_cross_entropy\'.")
        _sh_group.add_argument('--reduce-lr-on-plateau-factor', type=float, default=0.1,
                               help="Factor by which the learning rate will be reduced. new_lr = lr * factor. Default: 0.1.")
        _sh_group.add_argument('--reduce-lr-on-plateau-patience', type=int, default=5,
                               help="Number of epochs with no improvement after which LR will be reduced. Default: 5.")
        _sh_group.add_argument('--reduce-lr-on-plateau-min-lr', type=float, default=0,
                               help="Lower bound on the learning rate. Default: 0.")
        _sh_group.add_argument('--reduce-lr-every-nepochs-factor', type=float, default=0.1,
                               help="Factor by which the learning rate will be reduced. new_lr = lr * factor. Default: 0.1.")
        _sh_group.add_argument('--reduce-lr-every-nepochs-nepochs', type=int, default=5,
                               help="Number of epochs after which LR will be periodically reduced. Default: 5.")
        _sh_group.add_argument('--reduce-lr-every-nepochs-min-lr', type=float, default=0,
                               help="Lower bound on the learning rate. Default: 0.")
        _sh_group.add_argument('--stop-at-loss-metric', type=str, default="val_accuracy",
                               help="Loss metric monitored by stop_at_loss LR scheduler. Default: \'val_accuracy\'.")
        _sh_group.add_argument('--stop-at-loss-threshold', type=float, default=0,
                               help="Metric threshold monitored by stop_at_loss LR scheduler. Default: 0.")
        _sh_group.add_argument('--model-checkpoint-metric', type=str, default="val_categorical_cross_entropy",
                               help="Loss metric monitored by model_checkpoint LR scheduler. Default: \'val_categorical_cross_entropy\'")
        _sh_group.add_argument('--model-checkpoint-save-freq', type=int, default=2,
                               help="Frequency (in epochs) at which the model weights and bias will be saved by the model_checkpoint LR scheduler. Default: 2.")

        # Parallel execution options
        _pe_group = self.add_argument_group("Parallel execution options")
        _pe_group.add_argument('--parallel', type=str, default="sequential", choices=["sequential", "data"],
                               help="Data parallelization modes: \'sequential\', \'data\' (MPI). Default: \'sequential\'.")
        _pe_group.add_argument('--use-blocking-mpi', type=bool_lambda, default=True,
                               help="Enable non-blocking MPI primitives. Default: True.")
        _pe_group.add_argument('--use-mpi-buffers', type=bool_lambda, default=None,
                               help="Enable the use of MPI buffers. Possible values: \'True\' (MPI operations by buffer), \'False\' (MPI operations by object) or undefined (auto-select the better option). Default: undefined.")
        _pe_group.add_argument('--enable-gpu', type=bool_lambda, default=False,
                               help="Enable GPU, use cuDNN library. Default: False.")
        _pe_group.add_argument('--enable-gpudirect', type=bool_lambda, default=False,
                               help="Enable GPU pinned memory for gradients when using a CUDA-aware MPI version. Default: False.")
        _pe_group.add_argument('--enable-nccl', type=bool_lambda, default=False,
                               help="Enable the use of the NCCL library for  collective communications on GPUs. This option can only be set  with \'--enable-gpu\'. Default. False.")
        _pe_group.add_argument('--enable-cudnn-auto-conv-alg', type=bool_lambda, default=True,
                               help="Let cuDNN to select the best performing convolution algorithm. Default: True.")

        # Encryption options
        _cy_group = self.add_argument_group("Encryption options")
        _cy_group.add_argument('--encryption', dest="encryption_name", type=str, default="",
                               help="Encryption library: \'tenseal\', \'openfhe\', \'\' (None). Default \'\' (None).")
        _cy_group.add_argument('--encryption-poly-degree', type=int, default=13,
                               help="Encryption polynomial degree. 2 ^ \'value\'. Default: 13.")
        _cy_group.add_argument('--encryption-global-scale', type=int, default=40,
                               help="Encryption global scale. 2 ^ \'value'\'. Default: 40.")
        _cy_group.add_argument('--encryption-security-level', type=int, default=128,
                               help="Encryption security level: 0 (Not set), 128, 192, 256. Default: 128.")

        # Tracing and profiling
        _tr_group = self.add_argument_group("Tracing options")
        _tr_group.add_argument('--tracing', type=bool_lambda, default=False,
                               help="Obtain Simple/Extrae-based traces. Deffault: False.")
        _tr_group.add_argument('--tracer-output', type=str, default="",
                               help="Output file to store the Simple/Extrae-based traces.")
        _tr_group.add_argument('--tracer-pmlib-server', type=str, default="127.0.0.1",
                               help="Address of PMlib tracer server. Default: \'127.0.0.1\'.")
        _tr_group.add_argument('--tracer-pmlib-port', type=int, default=6526,
                               help="Port of PMlib tracer server. Default: 6526.")
        _tr_group.add_argument('--tracer-pmlib-device', type=str, default="",
                               help="Port of PMlib tracer device.")
        _tr_group.add_argument('--profile', type=bool_lambda, default=False,
                               help="Obtain cProfile profiles. Default: False.")

        # Performance modeling options
        _pm_group = self.add_argument_group("Performance modeling options")
        _pm_group.add_argument('--cpu-speed', type=float, default=4e12, help=argparse.SUPPRESS)
        _pm_group.add_argument('--memory-bw', type=float, default=50e9, help=argparse.SUPPRESS)
        _pm_group.add_argument('--network-bw', type=float, default=1e9, help=argparse.SUPPRESS)
        _pm_group.add_argument('--network-lat', type=float, default=0.5e-6, help=argparse.SUPPRESS)
        _pm_group.add_argument('--network-alg', type=str, default="vdg", choices=["bta", "vdg"], help=argparse.SUPPRESS)

        # Add Runtime parallel execution options
        _re_group = self.add_argument_group("Runtime parallel execution options")
        _re_group.add_argument('--mpi-processes', type=int, default=-1, help=argparse.SUPPRESS)
        _re_group.add_argument('--threads-per-process', type=int, default=-1, help=argparse.SUPPRESS)
        _re_group.add_argument('--gpus-per-node', type=int, default=-1, help=argparse.SUPPRESS)

        # Add Communication options
        _cm_group = self.add_argument_group("Communication options")
        _cm_group.add_argument('--mpi-protocol', type=str, default="", help=argparse.SUPPRESS)
        _cm_group.add_argument('--mpi-server', type=str, default="", help=argparse.SUPPRESS)
        _cm_group.add_argument('--mpi-port', type=int, default=-1, help=argparse.SUPPRESS)

    def parse_args(self, args=None, namespace=None):
        # Call super.parse_args
        result = super().parse_args(args, namespace)
        # Add runtime data
        result.mpi_processes = _get_mpi_processes()
        result.threads_per_process = _get_threads_per_process()
        result.gpus_per_node = _get_gpus_per_node()
        result.mpi_protocol = _get_mpi_protocol()
        result.mpi_server = _get_mpi_server()
        result.mpi_port = _get_mpi_port()
        # Populate self.lines (for self.print_args())
        if len(self.lines) == 0:
            lines = []
            for action_group in self._action_groups:
                indent = ""
                length = 0
                if action_group.title not in ('positional arguments', 'optional arguments'):
                    indent = "  "
                    lines.append("")
                    lines.append(action_group.title)
                    if action_group.description is not None:
                        lines.append(action_group.description)
                for action in action_group._group_actions:
                    if action.default == '==SUPPRESS==':
                        continue
                    option_string = f"{action.option_strings[0].replace('--', '')}"
                    if len(option_string) > length:
                        length = len(option_string)
                for action in action_group._group_actions:
                    if action.default == '==SUPPRESS==':
                        continue
                    option_string = f"{action.option_strings[0].replace('--', '')}"
                    tab = " " * (length - len(option_string))
                    lines.append(f"{indent}{option_string}{tab}: {getattr(result, action.dest)}")
            lines.append('')
            self.lines = lines
        return result

    def print_args(self) -> None:
        print("\n".join(self.lines))

    def get_default_values(self) -> dict[str, Any]:
        return vars(self.parse_args([]))

    def to_dict(self) -> dict[str, Any]:
        return vars(self.parse_args())



import pycuda.gpuarray as gpuarray  #type: ignore
from pycuda.driver import Function  #type: ignore

from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.losses.loss import Loss
from pydtnn.model import Model
from pydtnn.utils.constants import ArrayShape

class LossGPU(Loss[TensorGPU]):
    """
    Extends a Loss class with the attributes and methods required by GPU Losses.
    """

    def __init__(self, shape: ArrayShape, eps=1e-8):
        super().__init__(shape, eps)
        # NOTE: The following attributes will be initialized later.
        self.grid = None
        self.block = None

    def initialize(self) -> None:
        super().initialize()
        # NOTE: the model must be executed before this one.
        self.grid = self.model.cuda_grid
        self.block = self.model.cuda_block
        self.loss = gpuarray.empty((self.model.batch_size,), self.model.dtype)
        dx_gpu = gpuarray.empty(self.shape, self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)
        self.kernel = self.__init_gpu_kernel__()

    def __init_gpu_kernel__(self) -> Function:
        raise NotImplementedError()



from typing import TYPE_CHECKING

from numpy import ndarray

from pydtnn.schedulers.scheduler import Scheduler

if TYPE_CHECKING:
    from pydtnn.model import Model


class ReduceLREveryNEpochs(Scheduler):
    """
    ReduceLREveryNEpochs LRScheduler
    """

    def __init__(self, factor=0.1, nepochs=5, min_lr=0, verbose=True):
        super().__init__(verbose)
        self.factor = factor
        self.nepochs = nepochs
        self.min_lr = min_lr

    def on_epoch_end(self, train_loss: ndarray[float], val_loss: ndarray[float]) -> None:
        self.epoch_count += 1
        if self.epoch_count % self.nepochs == 0 and self.model.optimizer.learning_rate * self.factor >= self.min_lr:
            self.model.optimizer.learning_rate *= self.factor
            self.log(f"Setting learning rate to {self.model.optimizer.learning_rate:.8f}!")

    @classmethod
    def from_model(cls, model: "Model") -> "ReduceLREveryNEpochs":
        return ReduceLREveryNEpochs(model.reduce_lr_every_nepochs_factor,
                                    model.reduce_lr_every_nepochs_nepochs,
                                    model.reduce_lr_every_nepochs_min_lr)



from pydtnn.backends.cpu.utils.adaptive_avg_pooling_nchw_cython import adaptive_avg_pooling_bwd_nchw_cython, adaptive_avg_pooling_fwd_nchw_cython
from pydtnn.backends.cpu.utils.adaptive_avg_pooling_nhwc_cython import adaptive_avg_pooling_bwd_nhwc_cython, adaptive_avg_pooling_fwd_nhwc_cython
from pydtnn.layers.adaptive_average_pool_2d import AdaptiveAveragePool2D
from pydtnn.backends.cpu.layers.layer import LayerCPU

# Imports for the method from AbstractPool2DLayerCPU
from pydtnn.utils.tensor import TensorFormat

# Imports for the methods from AveragePool2DCPU
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
import numpy as np


class AdaptiveAveragePool2DCPU(AdaptiveAveragePool2D[np.ndarray], LayerCPU):
    # The backend is almost the same as a AveragePool2D layer.

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # The following atributte will be initalized in "initalize"
        self.y: np.ndarray = None  # type: ignore
    # -- END __init__ -- #

    # Method from AbstractPool2DLayerCPU
    def initialize(self, prev_shape: tuple[int, int], x: np.ndarray | None = None):
        # The objective is following lines is to override the AbstractPool2DLayer's initialize method, that is avoiding call to "super" since in that case AbstractPool2DLayer will be called eventually.
        super().initialize(prev_shape, x)

        match self.model.tensor_format:
            case TensorFormat.NCHW:
                self._forward = self._forward_nchw_cython
                self._backward = self._backward_nchw_cython
            case TensorFormat.NHWC:
                self._forward = self._forward_nhwc_cython
                self._backward = self._backward_nhwc_cython
            case _:
                raise NotImplementedError(f"\"AdaptiveAveragePool2DCPU\" is not implemented for \"{self.model.tensor_format}\" format.")

        y_shape = self.model.encode_shape((self.model.batch_size, self.co, self.ho, self.wo))
        # NOTE: This attribute only stores data, its value before the operation doesn't matter; it's initalized due avoid warnings in "LayerAndActivationBase.export".
        self.y = np.zeros(y_shape, dtype=self.model.dtype, order="C")

        if self.pooling_not_needed:
            self._forward = (lambda x: x)
        # else: Nothing special.

    # -- END initialize -- #

    def forward(self, x: np.ndarray) -> np.ndarray:
        return self._forward(x)

    def backward(self, dy: np.ndarray) -> np.ndarray:
        return self._backward(dy)

    def _forward_nhwc_cython(self, x: np.ndarray) -> np.ndarray:
        y = self.y[:x.shape[0], :]
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_ADP_AVG_POOL)
        adaptive_avg_pooling_fwd_nhwc_cython(x, y)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def _forward_nchw_cython(self, x: np.ndarray) -> np.ndarray:
        y = self.y[:x.shape[0], :]
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_ADP_AVG_POOL)
        adaptive_avg_pooling_fwd_nchw_cython(x, y)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def _backward_nhwc_cython(self, dy: np.ndarray) -> np.ndarray:
        dx = np.zeros((dy.shape[0], self.hi, self.wi, self.ci), dtype=self.model.dtype, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_ADP_AVG_POOL)
        adaptive_avg_pooling_bwd_nhwc_cython(dy, dx)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(dx, dtype=self.model.dtype, order='C', copy=None)

    def _backward_nchw_cython(self, dy: np.ndarray) -> np.ndarray:
        dx = np.zeros((dy.shape[0], self.ci, self.hi, self.wi), dtype=self.model.dtype, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_ADP_AVG_POOL)
        adaptive_avg_pooling_bwd_nchw_cython(dy, dx)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(dx, dtype=self.model.dtype, order='C', copy=None)



#!/usr/bin/env python

"""
PyDTNN Benchmark script
"""

import cProfile
import os
from pathlib import Path
import sys
import time
from datetime import datetime

from pydtnn.model import Model
from pydtnn.utils import random
from pydtnn.parser import PydtnnArgumentParser
from pydtnn.utils.best_of import BestOf

ompi_stdout_rank = os.environ.get("OMPI_STDOUT_RANK", None)
if ompi_stdout_rank and os.environ.get("OMPI_COMM_WORLD_RANK", "0") != ompi_stdout_rank:
    sys.stdout = sys.stderr = open(os.devnull, "w")

Extrae_tracing = False
if os.environ.get("EXTRAE_ON", None) == "1":
    TracingLibrary = "libptmpitrace.so"
    import pyextrae.common.extrae as pyextrae

    pyextrae.startTracing(TracingLibrary)
    Extrae_tracing = True


def show_options(params):
    for arg in vars(params):
        if arg != "comm":
            print(f'  {arg:31s}: {str(getattr(params, arg)):s}')
            # print(f'  --{arg:s}={str(getattr(params, arg)):s} \\')


def print_model_reports(model):
    # Print performance counter report
    model.perf_counter.print_report()
    # Print BestOf report
    if model.enable_best_of:
        print()
        BestOf.print_report()


def main():
    # Parse options
    parser = PydtnnArgumentParser()
    config = parser.parse_args()
    # Initialize random seed
    random.seed(config.random_seed)
    # Create model
    model = Model(**vars(config))
    # Print model
    if model.comm_rank == 0:
        model.show_model()
        print()
        model.show_layers()
        print()
    # Print parameters
    if model.comm_rank == 0:
        print('**** Parameters:')
        parser.print_args()
    # First (or unique) evaluation
    if model.evaluate_on_train or model.evaluate_only:
        if model.comm_rank == 0:
            print('**** Evaluating on test dataset...')
            t1 = time.time()
        _ = model.evaluate_dataset()
        if model.comm_rank == 0:
            t2 = time.time()
            # noinspection PyUnboundLocalVariable
            total_time = t2 - t1
            if model.evaluate_only:
                print(f'Testing time: {total_time:5.4f} s')
                print(f'Testing throughput: {model.dataset.test_nsamples / total_time:5.4f} samples/s')
        if model.evaluate_only:
            print_model_reports(model)
            raise SystemExit(0)
    # Barrier
    if model.parallel in ["data"]:
        model.comm.Barrier()
    # Training
    if model.comm_rank == 0:
        # print('**** Model time: ', model.calculate_time())
        print('**** Training...')
        t1 = time.time()
        if model.profile:
            pr = cProfile.Profile()
            pr.enable()
    # Training a model directly from a dataset
    # or alternatively, define any custom data
    # mode.dataset = CustomDataset(model, x, y)
    history = model.train_dataset()
    # Barrier
    if model.parallel == "data":
        model.comm.Barrier()
    # Print performance results and evaluation history
    if model.comm_rank == 0:
        if model.profile:
            pr.disable()
            stamp = datetime.now().isoformat(timespec="seconds").replace(" ", "-").replace(":", "-").replace(".", "-")
            stats = Path(f"profile-{stamp}.stat").resolve()
            pr.dump_stats(stats)
            print(f'Dumped profile stats to: {stats}')
        t2 = time.time()
        print('**** Done...')
        total_time = t2 - t1
        print(f'Training and validation time: {total_time:5.4f} s')
        if model.perf_counter.num_epochs > 0:
            print(f'Training and validation time per epoch: {total_time / model.perf_counter.num_epochs:5.4f} s')
            print(f'Training and validation throughput: '
                  f'{(model.dataset.train_nsamples * model.perf_counter.num_epochs) / total_time:5.4f} samples/s')
        if model.history_file:
            with open(model.history_file, "w") as f:
                epochs = max(len(v) for v in history.values())
                for epoch in range(epochs):
                    f.write(f"epoch: {epoch}\n")
                    for key in history:
                        f.write(f"    {key}: {history[key][epoch]}\n")
    # Second (and last) evaluation
    if model.evaluate_on_train:
        if model.comm_rank == 0:
            print('**** Evaluating on test dataset...')
            t1 = time.time()
        _ = model.evaluate_dataset()
        if model.comm_rank == 0:
            t2 = time.time()
            # noinspection PyUnboundLocalVariable
            total_time = t2 - t1
            if not model.evaluate_only:
                print(f'Testing time: {total_time:5.4f} s')
                print(f'Testing throughput: {model.dataset.test_nsamples / total_time:5.4f} samples/s')
    # Print model reports
    if model.comm_rank == 0:
        print_model_reports(model)
    # Barrier and finalize
    if model.comm is not None and model.MPI is not None:
        model.comm.Barrier()
        # The next line is required if running under SLURM (it seems it is not automatically called at exit)
        model.MPI.Finalize()


if __name__ == "__main__":
    main()



# _______________________________________________________________________________________________________________
# In this file must be implemented only the translation of PyTorch Convolutional layers to its PyDTNN equivalent.
# _______________________________________________________________________________________________________________

# Typing related (or non important) imports
from typing import Dict

# Functionality imports
from pydtnn.layers.flatten import Flatten as _Flatten
# import pydtnn.converters.pytorch2pydtnn.common as cm

# ------------------ #


def Flatten(args: Dict[str, str]) -> _Flatten:
    # https://pytorch.org/docs/stable/generated/torch.nn.Flatten.html#torch.nn.Flatten
    # torch.nn.Flatten(start_dim=1, end_dim=-1)

    # PyTorch attributes:
    # Not used: start_dim, end_dim (It's not used due the way the layer's initialization works in PyDTNN)
    # ---- #
    # PyDTNN attributes: None
    # ---- #
    not_used = args

    return _Flatten()
# ------------------ #



# _______________________________________________________________________________________________________________
# In this file must be implemented only the translation of PyTorch functions to its PyDTNN equivalent.
# _______________________________________________________________________________________________________________

# Typing related (or non important) imports
from typing import Dict, Any, Tuple, List
from pydtnn.layer_base import LayerBase

# Functionality imports
from pydtnn.layers.average_pool_2d import AveragePool2D
from pydtnn.layers.adaptive_average_pool_2d import AdaptiveAveragePool2D
from pydtnn.layers.addition_block import AdditionBlock
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.flatten import Flatten
from pydtnn.activations.log import Log
from pydtnn.activations.relu import Relu
from pydtnn.activations.sigmoid import Sigmoid
from pydtnn.activations.softmax import Softmax
from pydtnn.activations.tanh import Tanh
import pydtnn.converters.pytorch2pydtnn.common as cm
from pydtnn.converters.pytorch2pydtnn.layers import activation

# ------------------ #
# - Torch Functions  #
# ------------------ #


def adaptive_avg_pool_2d(args: Dict[str, str]) -> Tuple[AveragePool2D, str]:
    # It is not the layer, but the operation itself.
    # from torch.nn.functional import adaptive_avg_pool2d
    # adaptive_avg_pool2d(input: Tensor, output_size: BroadcastingList2[int])

    dict_params = dict()
    # Example: torch.nn.functional.adaptive_avg_pool2d(relu, (1, 1)) | args = 'relu, (1, 1)'
    params: List[str] = args[cm.PARAMETERS].split(cm.ARGS_SEPARATOR)
    # removing the input layer:
    dict_params["input"] = params.pop(0)  # Situation after operation: [] or ['number'] or ['(number', 'number)']

    # Getting the arguments:
    match len(params):
        case 0:
            params = None
        case 1:
            param = int(params[0])
            params = [param, param]  # Only 1 argument implies the weight and height are the same.
        case greater_than_1:  # len must be always >= 0
            params = [int(param.replace('(', '').replace(')', '')) for param in params]

    if params is not None:
        dict_params[cm.ARGUMENTS] = {cm.PYTORCH_OUTPUT_SIZE: params}

    return (AdaptiveAveragePool2D(dict_params), dict_params["input"])


def add(args: Dict[str, Any]) -> Tuple[AdditionBlock, str]:
    # https://pytorch.org/docs/stable/generated/torch.add.html

    # It should be prepared so the params have the following format: "[layer1,layer2]"
    layer_name: str = args[cm.OPERATION_VAR]
    dict_equivalent_layers = args[cm.EQUIVALENT_LAYERS]
    params = cm.separate_function_params(args[cm.PARAMETERS])

    params = cm.get_equivalent_layer(params, dict_equivalent_layers)
    dict_layers: Dict[str, Tuple[LayerBase, str]] = args[cm.LAYERS]

    list_layers, to_remove, input_layer_name = cm.get_lists_operations_and_outputs(dict_layers=dict_layers, layer_inputs=params)

    to_remove = set(to_remove)  # Remove multiple ocurrences of a layer. Consecuence of "get_equivalent_layer".
    # The removed layers will be accesed through the AdditionBlock.
    for elem in to_remove:
        del dict_layers[elem]
    # The equivalences dictionary values are set
    for elem in params:
        dict_equivalent_layers[elem] = layer_name
    # NOTE: Not always "params == to_remove"
    for elem in to_remove:
        dict_equivalent_layers[elem] = layer_name

    # AdditionBlock expects every "branch" (layer list) as a different argument.
    return (AdditionBlock(*list_layers), input_layer_name)


def concat(args: Dict[str, Any]) -> Tuple[ConcatenationBlock, str]:
    # https://pytorch.org/docs/main/generated/torch.cat.html

    # TODO: es necesario hacer un diccionario que sustituya los parámetros que ya han sido introducidos por la capa de concatenación/adición.
    # También hay que haer que solo aparezca una única vez.
    layer_name: str = args[cm.OPERATION_VAR]
    dict_equivalent_layers: Dict[str, str] = args[cm.EQUIVALENT_LAYERS]
    parameters: List[str] = args[cm.PARAMETERS].split("],")

    params = parameters.pop(0)  # Since PyDTNN always concatenate in the same dimensions, the rest of the PyTorch parameters can be ignored
    params = cm.separate_function_params(params)
    params = cm.get_equivalent_layer(params, dict_equivalent_layers)

    dict_layers: Dict[str, Tuple[LayerBase, str]] = args[cm.LAYERS]
    list_layers, to_remove, input_layer_name = cm.get_lists_operations_and_outputs(dict_layers=dict_layers, layer_inputs=params)

    to_remove = set(to_remove)  # Remove multiple ocurrences of a layer. Consecuence of "get_equivalent_layer".

    # The removed layers will be accesed through the ConcatenationBlock.
    for elem in to_remove:
        del dict_layers[elem]
    # The equivalences dictionary values are set
    for elem in params:
        dict_equivalent_layers[elem] = layer_name
    # NOTE: Not always "params == to_remove"
    for elem in to_remove:
        dict_equivalent_layers[elem] = layer_name

    # ConcatenationBlock expects every "branch" (layer list) as a different argument.
    return (ConcatenationBlock(*list_layers), input_layer_name)


def flatten(args: Dict[str, str]) -> Tuple[Flatten, str]:
    # https://pytorch.org/docs/stable/generated/torch.flatten.html
    # torch.flatten(input, start_dim=0, end_dim=-1)

    def switch(list_params: List[str], dict_params: Dict[str, str] = dict()) -> Dict[str, str]:
        # This is a switch with "fall through".
        match len(list_params):
            case 3:
                var = list_params.pop().split("end_dim=")
                dict_params["end_dim"] = int(var.pop())
                # // fall through
                return switch(list_params, dict_params)
            case 2:
                var = list_params.pop().split("start_dim=")
                dict_params["start_dim"] = int(var.pop())
                # // fall through
                return switch(list_params, dict_params)
            case 1:
                dict_params["input"] = list_params.pop()
                # // fall through
                return switch(list_params, dict_params)
            case _:
                return dict_params
    # - END switch - #
    params = args[cm.PARAMETERS].strip()
    dict_params = switch(params.split(cm.ARGS_SEPARATOR))

    return (Flatten(dict_params), dict_params["input"])
# ------------------ #

# ------------------ #
# --- Activations -- #
# ------------------ #


def log(args: Dict[str, Any]) -> Tuple[Log, str]:
    # https://pytorch.org/docs/stable/generated/torch.nn.functional.logsigmoid.html#torch.nn.functional.logsigmoid

    dict_params = dict()

    # Example: torch.nn.functional.relu(features_norm5, inplace = True)
    params = args[cm.PARAMETERS].strip().split("inplace=")
    inplace = bool(params.pop()) if len(params) > 0 else None

    dict_params[cm.ARGUMENTS] = {"input": params[0].split(cm.ARGS_SEPARATOR)[0]}
    if inplace is not None:
        dict_params["inplace"] = inplace

    return (activation.LogSigmoid(**dict_params), dict_params["input"])


def relu(args: Dict[str, str]) -> Tuple[Relu, str]:

    # https://pytorch.org/docs/stable/generated/torch.nn.functional.relu.html#torch.nn.functional.relu
    # It is not the layer, but the operation itself.
    # from torch.nn.functional import relu
    # relu(input: Tensor, inplace: bool = False)

    dict_params = dict()

    # Example: torch.nn.functional.relu(features_norm5, inplace = True)
    params = args[cm.PARAMETERS].strip().split("inplace=")
    inplace = bool(params.pop()) if len(params) > 0 else None

    dict_params[cm.ARGUMENTS] = {"input": params[0].split(cm.ARGS_SEPARATOR)[0]}
    if inplace is not None:
        dict_params["inplace"] = inplace

    return (activation.ReLU(dict_params), dict_params[cm.ARGUMENTS]["input"])


def sigmoid(args: Dict[str, Any]) -> Tuple[Sigmoid, str]:
    # https://pytorch.org/docs/stable/generated/torch.nn.functional.sigmoid.html#torch.nn.functional.sigmoid
    # Not used Pytorch's parameters: inplace.

    dict_params = dict()

    params: List[str] = args[cm.PARAMETERS].split(cm.ARGS_SEPARATOR)
    # removing the input layer:
    dict_params["input"] = params.pop(0)

    return (activation.Sigmoid(**dict_params), dict_params["input"])


def softmax(args: Dict[str, Any]) -> Tuple[Softmax, str]:
    # https://pytorch.org/docs/stable/generated/torch.nn.functional.softmax.html#torch.nn.functional.softmax
    # softmax(input, dim=None, _stacklevel=3, dtype=None)

    def switch(list_params: List[str], dict_params: Dict[str, str] = dict()) -> Dict[str, str]:
        # This is a switch with "fall through".
        match len(list_params):
            case 3:
                var = list_params.pop().split("dim=")
                dict_params["end_dim"] = int(var.pop())
                # // fall through
                return switch(list_params, dict_params)
            case 2:
                var = list_params.pop().split("dtype=")
                dict_params["start_dim"] = int(var.pop())
                # // fall through
                return switch(list_params, dict_params)
            case 1:
                dict_params["input"] = list_params.pop()
                # // fall through
                return switch(list_params, dict_params)
            case _:
                return dict_params
    # - END switch - #
    params = args[cm.PARAMETERS].strip()
    dict_params = switch(params.split(cm.ARGS_SEPARATOR))

    return (activation.Softmax(**dict_params), dict_params["input"])


def tanh(args: Dict[str, Any]) -> Tuple[Tanh, str]:
    # https://pytorch.org/docs/stable/generated/torch.nn.functional.tanh.html#torch.nn.functional.tanh
    dict_params = dict()

    params: List[str] = args[cm.PARAMETERS].split(cm.ARGS_SEPARATOR)
    # removing the input layer:
    dict_params["input"] = params.pop(0)

    return (activation.Tanh(**dict_params), dict_params["input"])

# ------------------ #
